import dataclasses
import os
import platform
import shlex
import shutil
import subprocess
import sys
import textwrap
from collections.abc import Sequence, Set
from pathlib import Path
from typing import assert_never

from filelock import FileLock

from cibuildwheel import errors
from cibuildwheel.architecture import Architecture
from cibuildwheel.environment import ParsedEnvironment
from cibuildwheel.frontend import BuildFrontendName, get_build_frontend_extra_flags
from cibuildwheel.logger import log
from cibuildwheel.options import Options
from cibuildwheel.platforms.macos import install_cpython as install_build_cpython
from cibuildwheel.selector import BuildSelector
from cibuildwheel.util import resources
from cibuildwheel.util.cmd import call, shell, split_command
from cibuildwheel.util.file import CIBW_CACHE_PATH, copy_test_sources, download, move_file
from cibuildwheel.util.helpers import prepare_command, unwrap_preserving_paragraphs
from cibuildwheel.util.packaging import find_compatible_wheel
from cibuildwheel.venv import constraint_flags, virtualenv


@dataclasses.dataclass(frozen=True, kw_only=True)
class PythonConfiguration:
    version: str
    identifier: str
    url: str
    build_url: str

    @property
    def sdk(self) -> str:
        return self.multiarch.rsplit("-", 1)[1]

    @property
    def arch(self) -> str:
        return self.multiarch.rsplit("-", 1)[0]

    @property
    def multiarch(self) -> str:
        # The multiarch identifier, as reported by `sys.implementation._multiarch`
        return "-".join(self.identifier.split("-ios_")[1].rsplit("_", 1))

    @property
    def is_simulator(self) -> bool:
        return self.identifier.endswith("_iphonesimulator")

    @property
    def xcframework_slice(self) -> str:
        "XCframeworks include binaries for multiple ABIs; which ABI section should be used?"
        return "ios-arm64_x86_64-simulator" if self.is_simulator else "ios-arm64"


def all_python_configurations() -> list[PythonConfiguration]:
    # iOS builds are always cross builds; we need to install a macOS Python as
    # well. Rather than duplicate the location of the URL of macOS installers,
    # load the macos configurations, determine the macOS configuration that
    # matches the platform we're building, and embed that URL in the parsed iOS
    # configuration.
    macos_python_configs = resources.read_python_configs("macos")

    def build_url(config_dict: dict[str, str]) -> str:
        # The iOS identifier will be something like cp313-ios_arm64_iphoneos.
        # Drop the iphoneos suffix, then replace ios with macosx to yield
        # cp313-macosx_arm64, which will be a macOS build identifier.
        modified_ios_identifier = config_dict["identifier"].rsplit("_", 1)[0]
        macos_identifier = modified_ios_identifier.replace("ios", "macosx")
        matching = [
            config for config in macos_python_configs if config["identifier"] == macos_identifier
        ]
        return matching[0]["url"]

    # Load the platform configuration
    full_python_configs = resources.read_python_configs("ios")
    # Build the configurations, annotating with macOS URL details.
    return [
        PythonConfiguration(
            **item,
            build_url=build_url(item),
        )
        for item in full_python_configs
    ]


def get_python_configurations(
    build_selector: BuildSelector,
    architectures: Set[Architecture],
) -> list[PythonConfiguration]:
    python_configurations = all_python_configurations()

    # Filter out configs that don't match any of the selected architectures
    python_configurations = [
        c
        for c in python_configurations
        if any(c.identifier.endswith(f"-ios_{a.value}") for a in architectures)
    ]

    # Skip builds as required by BUILD/SKIP
    python_configurations = [c for c in python_configurations if build_selector(c.identifier)]

    return python_configurations


def install_target_cpython(tmp: Path, config: PythonConfiguration, free_threading: bool) -> Path:
    if free_threading:
        msg = "Free threading builds aren't available for iOS (yet)"
        raise errors.FatalError(msg)

    # Install an iOS build of CPython
    ios_python_tar_gz = config.url.rsplit("/", 1)[-1]
    extension = ".tar.gz"
    assert ios_python_tar_gz.endswith(extension)
    installation_path = CIBW_CACHE_PATH / ios_python_tar_gz[: -len(extension)]
    with FileLock(str(installation_path) + ".lock"):
        if not installation_path.exists():
            downloaded_tar_gz = tmp / ios_python_tar_gz
            download(config.url, downloaded_tar_gz)
            installation_path.mkdir(parents=True, exist_ok=True)
            call("tar", "-C", installation_path, "-xf", downloaded_tar_gz)
            downloaded_tar_gz.unlink()

    return installation_path


def cross_virtualenv(
    *,
    py_version: str,
    target_python: Path,
    multiarch: str,
    build_python: Path,
    venv_path: Path,
    dependency_constraint: Path | None,
    xbuild_tools: Sequence[str] | None,
) -> dict[str, str]:
    """Create a cross-compilation virtual environment.

    In a cross-compilation environment, the *target* is the platform where the
    code will ultimately run, and the *build* is the platform where you're
    running the compilation. When building iOS wheels, iOS is the target machine
    and macOS is the build machine. The terminology around these machines varies
    between build tools (configure uses "host" and "build"; cmake uses "target" and
    "build host").

    A cross-compilation virtualenv is an environment that is based on the
    *build* python (so that binaries can execute); but it modifies the
    environment at startup so that any request for platform details (such as
    `sys.platform` or `sysconfig.get_platform()`) return details of the target
    platform. It also applies a loader patch so that any virtualenv created by
    the cross-compilation environment will also be a cross-compilation
    environment.

    :param py_version: The Python version (major.minor) in use
    :param target_python: The path to the python binary for the target platform
    :param multiarch: The multiarch tag for the target platform (i.e., the value
        of `sys.implementation._multiarch`)
    :param build_python: The path to the python binary for the build platform
    :param venv_path: The path where the cross virtual environment should be
        created.
    :param dependency_constraint: A path to a constraint file that should be
        used when constraining dependencies in the environment.
    :param xbuild_tools: A list of executable names (without paths) that are
        on the path, but must be preserved in the cross environment.
    """
    # Create an initial macOS virtual environment
    env = virtualenv(
        py_version,
        build_python,
        venv_path,
        dependency_constraint,
        use_uv=False,
    )

    # Convert the macOS virtual environment into an iOS virtual environment
    # using the cross-platform conversion script in the iOS distribution.

    # target_python is the path to the Python binary;
    # determine the root of the XCframework slice that is being used.
    slice_path = target_python.parent.parent
    call(
        "python",
        str(slice_path / f"platform-config/{multiarch}/make_cross_venv.py"),
        str(venv_path),
        env=env,
        cwd=venv_path,
    )

    # When running on macOS, it's easy for the build environment to leak into
    # the target environment, especially when building for ARM64 (because the
    # build architecture is the same as the target architecture). The primary
    # culprit for this is Homebrew libraries leaking in as dependencies for iOS
    # libraries.
    #
    # To prevent problems, set the PATH to isolate the build environment from
    # sources that could introduce incompatible binaries.
    #
    # However, there may be some tools on the path that are needed for the
    # build. Find their location on the path, and link the underlying binaries
    # (fully resolving symlinks) to a "safe" location that will *only* contain
    # those tools. This avoids needing to add *all* of Homebrew to the path just
    # to get access to (for example) cmake for build purposes. A value of None
    # means the user hasn't provided a list of xbuild tools.
    xbuild_tools_path = venv_path / "cibw_xbuild_tools"
    xbuild_tools_path.mkdir()
    if xbuild_tools is None:
        log.warning(
            textwrap.dedent(
                """
                Your project configuration does not define any cross-build tools.

                iOS builds use an isolated build environment; if your build process requires any
                third-party tools (such as cmake, ninja, or rustc), you must explicitly declare
                that those tools are required using xbuild-tools/CIBW_XBUILD_TOOLS. This will
                likely manifest as a "somebuildtool: command not found" error.

                If the build succeeds, you can silence this warning by setting adding
                `xbuild-tools = []` to your pyproject.toml configuration, or exporting
                CIBW_XBUILD_TOOLS as an empty string into your environment.
                """
            )
        )
    else:
        for tool in xbuild_tools:
            tool_path = shutil.which(tool)
            if tool_path is None:
                msg = f"Could not find a {tool!r} executable on the path."
                raise errors.FatalError(msg)

            # Link the binary into the safe tools directory
            original = Path(tool_path).resolve()
            print(f"{tool!r} will be included in the cross-build environment (using {original})")
            (xbuild_tools_path / tool).symlink_to(original)

    env["PATH"] = os.pathsep.join(
        [
            # The target python's binary directory
            str(target_python.parent),
            # The cross-platform environment's binary directory
            str(venv_path / "bin"),
            # The directory of cross-build tools
            str(xbuild_tools_path),
            # The bare minimum Apple system paths.
            "/usr/bin",
            "/bin",
            "/usr/sbin",
            "/sbin",
            "/Library/Apple/usr/bin",
        ]
    )
    # Also unset DYLD_LIBRARY_PATH to ensure that no macOS libraries will be
    # found and linked.
    env.pop("DYLD_LIBRARY_PATH", None)

    return env


def setup_python(
    tmp: Path,
    *,
    python_configuration: PythonConfiguration,
    dependency_constraint: Path | None,
    environment: ParsedEnvironment,
    build_frontend: BuildFrontendName,
    xbuild_tools: Sequence[str] | None,
) -> tuple[Path, dict[str, str]]:
    # Not using set because mypy can't narrow it
    if build_frontend == "build[uv]" or build_frontend == "uv":  # noqa: PLR1714
        msg = "uv doesn't support iOS"
        raise errors.FatalError(msg)

    # An iOS environment requires 2 python installs - one for the build machine
    # (macOS), and one for the target (iOS). We'll only ever interact with the
    # *target* python, but the build Python needs to exist to act as the base
    # for a cross venv.
    tmp.mkdir()
    implementation_id = python_configuration.identifier.split("-")[0]
    log.step(f"Installing Build Python {implementation_id}...")
    if implementation_id.startswith("cp"):
        free_threading = "t-ios" in python_configuration.identifier
        build_python = install_build_cpython(
            tmp,
            python_configuration.version,
            python_configuration.build_url,
            free_threading,
        )
    else:
        msg = f"Unknown Python implementation: {implementation_id}"
        raise errors.FatalError(msg)

    assert build_python.exists(), (
        f"{build_python.name} not found, has {list(build_python.parent.iterdir())}"
    )

    log.step(f"Installing Target Python {implementation_id}...")
    target_install_path = install_target_cpython(tmp, python_configuration, free_threading)
    target_python = (
        target_install_path
        / "Python.xcframework"
        / python_configuration.xcframework_slice
        / "bin"
        / f"python{python_configuration.version}"
    )

    assert target_python.exists(), (
        f"{target_python.name} not found, has {list(target_install_path.iterdir())}"
    )

    log.step("Creating cross build environment...")

    venv_path = tmp / "venv"
    env = cross_virtualenv(
        py_version=python_configuration.version,
        target_python=target_python,
        multiarch=python_configuration.multiarch,
        build_python=build_python,
        venv_path=venv_path,
        dependency_constraint=dependency_constraint,
        xbuild_tools=xbuild_tools,
    )
    venv_bin_path = venv_path / "bin"
    assert venv_bin_path.exists()

    # We version pip ourselves, so we don't care about pip version checking
    env["PIP_DISABLE_PIP_VERSION_CHECK"] = "1"

    # upgrade pip to the version matching our constraints
    # if necessary, reinstall it to ensure that it's available on PATH as 'pip'
    pip = ["python", "-m", "pip"]
    call(
        *pip,
        "install",
        "--upgrade",
        "pip",
        *constraint_flags(dependency_constraint),
        env=env,
        cwd=venv_path,
    )

    # Apply our environment after pip is ready
    env = environment.as_dictionary(prev_environment=env)

    # Check what Python version we're on
    which_python = call("which", "python", env=env, capture_stdout=True).strip()
    print(which_python)
    if which_python != str(venv_bin_path / "python"):
        msg = (
            "cibuildwheel: python available on PATH doesn't match our installed instance. "
            "If you have modified PATH, ensure that you don't overwrite cibuildwheel's "
            "entry or insert python above it."
        )
        raise errors.FatalError(msg)
    call("python", "--version", env=env)

    # Check what pip version we're on
    assert (venv_bin_path / "pip").exists()
    which_pip = call("which", "pip", env=env, capture_stdout=True).strip()
    print(which_pip)
    if which_pip != str(venv_bin_path / "pip"):
        msg = (
            "cibuildwheel: pip available on PATH doesn't match our installed instance. "
            "If you have modified PATH, ensure that you don't overwrite cibuildwheel's "
            "entry or insert pip above it."
        )
        raise errors.FatalError(msg)
    call("pip", "--version", env=env)

    # Ensure that IPHONEOS_DEPLOYMENT_TARGET is set in the environment
    env.setdefault("IPHONEOS_DEPLOYMENT_TARGET", "13.0")

    log.step("Installing build tools...")
    match build_frontend:
        case "pip":
            # No additional build tools required
            pass
        case "build":
            call(
                "pip",
                "install",
                "--upgrade",
                "build[virtualenv]",
                *constraint_flags(dependency_constraint),
                env=env,
            )
        case _:
            assert_never(build_frontend)

    return target_install_path, env


def build(options: Options, tmp_path: Path) -> None:
    if sys.platform != "darwin":
        msg = "iOS binaries can only be built on macOS"
        raise errors.FatalError(msg)

    python_configurations = get_python_configurations(
        build_selector=options.globals.build_selector,
        architectures=options.globals.architectures,
    )

    if not python_configurations:
        return

    try:
        before_all_options_identifier = python_configurations[0].identifier
        before_all_options = options.build_options(before_all_options_identifier)

        if before_all_options.before_all:
            log.step("Running before_all...")
            env = before_all_options.environment.as_dictionary(prev_environment=os.environ)
            env.setdefault("IPHONEOS_DEPLOYMENT_TARGET", "13.0")
            before_all_prepared = prepare_command(
                before_all_options.before_all,
                project=".",
                package=before_all_options.package_dir,
            )
            shell(before_all_prepared, env=env)

        built_wheels: list[Path] = []

        for config in python_configurations:
            build_options = options.build_options(config.identifier)
            build_frontend = build_options.build_frontend
            # uv doesn't support iOS
            # Not using set because mypy can't narrow it
            if build_frontend.name == "build[uv]" or build_frontend.name == "uv":  # noqa: PLR1714
                msg = "uv doesn't support iOS"
                raise errors.FatalError(msg)

            log.build_start(config.identifier)

            identifier_tmp_dir = tmp_path / config.identifier
            identifier_tmp_dir.mkdir()
            built_wheel_dir = identifier_tmp_dir / "built_wheel"
            repaired_wheel_dir = identifier_tmp_dir / "repaired_wheel"

            constraints_path = build_options.dependency_constraints.get_for_python_version(
                version=config.version, tmp_dir=identifier_tmp_dir
            )

            target_install_path, env = setup_python(
                identifier_tmp_dir / "build",
                python_configuration=config,
                dependency_constraint=constraints_path,
                environment=build_options.environment,
                build_frontend=build_frontend.name,
                xbuild_tools=build_options.xbuild_tools,
            )

            compatible_wheel = find_compatible_wheel(built_wheels, config.identifier)
            if compatible_wheel:
                log.step_end()
                print(
                    f"\nFound previously built wheel {compatible_wheel.name} "
                    f"that is compatible with {config.identifier}. "
                    "Skipping build step..."
                )
                test_wheel = compatible_wheel
            else:
                if build_options.before_build:
                    log.step("Running before_build...")
                    before_build_prepared = prepare_command(
                        build_options.before_build,
                        project=".",
                        package=build_options.package_dir,
                    )
                    shell(before_build_prepared, env=env)

                log.step("Building wheel...")
                built_wheel_dir.mkdir()

                extra_flags = get_build_frontend_extra_flags(
                    build_frontend,
                    build_options.build_verbosity,
                    build_options.config_settings,
                    py38=False,
                )

                match build_frontend.name:
                    case "pip":
                        # Path.resolve() is needed. Without it pip wheel may try to
                        # fetch package from pypi.org. See
                        # https://github.com/pypa/cibuildwheel/pull/369
                        call(
                            "python",
                            "-m",
                            "pip",
                            "wheel",
                            build_options.package_dir.resolve(),
                            f"--wheel-dir={built_wheel_dir}",
                            "--no-deps",
                            *extra_flags,
                            env=env,
                        )
                    case "build":
                        call(
                            "python",
                            "-m",
                            "build",
                            build_options.package_dir,
                            "--wheel",
                            f"--outdir={built_wheel_dir}",
                            *extra_flags,
                            env=env,
                        )
                    case _:
                        assert_never(build_frontend)

                built_wheel = next(built_wheel_dir.glob("*.whl"))

                if built_wheel.name.endswith("none-any.whl"):
                    raise errors.NonPlatformWheelError()

                repaired_wheel_dir.mkdir()
                if build_options.repair_command:
                    log.step("Repairing wheel...")

                    repair_command_prepared = prepare_command(
                        build_options.repair_command,
                        wheel=built_wheel,
                        dest_dir=repaired_wheel_dir,
                        package=build_options.package_dir,
                        project=".",
                    )
                    shell(repair_command_prepared, env=env)
                else:
                    shutil.move(str(built_wheel), repaired_wheel_dir)

                try:
                    repaired_wheel = next(repaired_wheel_dir.glob("*.whl"))
                except StopIteration:
                    raise errors.RepairStepProducedNoWheelError() from None

                if repaired_wheel.name in {wheel.name for wheel in built_wheels}:
                    raise errors.AlreadyBuiltWheelError(repaired_wheel.name)

                test_wheel = repaired_wheel

                log.step_end()

            if build_options.test_command and build_options.test_selector(config.identifier):
                if not config.is_simulator:
                    log.step("Skipping tests on non-simulator SDK")
                elif config.arch != os.uname().machine:
                    log.step("Skipping tests on non-native simulator architecture")
                else:
                    test_env = build_options.test_environment.as_dictionary(prev_environment=env)

                    if build_options.before_test:
                        before_test_prepared = prepare_command(
                            build_options.before_test,
                            project=".",
                            package=build_options.package_dir,
                        )
                        shell(before_test_prepared, env=test_env)

                    log.step("Setting up test harness...")
                    # Clone the testbed project into the build directory
                    testbed_path = identifier_tmp_dir / "testbed"
                    call(
                        "python",
                        target_install_path / "testbed",
                        "clone",
                        testbed_path,
                        env=test_env,
                    )

                    testbed_app_path = testbed_path / "iOSTestbed" / "app"

                    # Copy the test sources to the testbed app
                    if build_options.test_sources:
                        copy_test_sources(
                            build_options.test_sources,
                            Path.cwd(),
                            testbed_app_path,
                        )
                    else:
                        (testbed_app_path / "test_fail.py").write_text(
                            resources.TEST_FAIL_CWD_FILE.read_text()
                        )

                    log.step("Installing test requirements...")
                    # Install the compiled wheel (with any test extras), plus
                    # the test requirements. Use the --platform tag to force
                    # the installation of iOS wheels; this requires the use of
                    # --only-binary=:all:
                    ios_version = test_env["IPHONEOS_DEPLOYMENT_TARGET"]
                    platform_tag = f"ios_{ios_version.replace('.', '_')}_{config.arch}_{config.sdk}"

                    call(
                        "python",
                        "-m",
                        "pip",
                        "install",
                        "--only-binary=:all:",
                        "--platform",
                        platform_tag,
                        "--target",
                        testbed_path / "iOSTestbed" / "app_packages",
                        f"{test_wheel}{build_options.test_extras}",
                        *build_options.test_requires,
                        env=test_env,
                    )

                    log.step("Running test suite...")

                    # iOS doesn't support placeholders in the test command,
                    # because the source dir isn't visible on the simulator.
                    if (
                        "{project}" in build_options.test_command
                        or "{package}" in build_options.test_command
                    ):
                        msg = unwrap_preserving_paragraphs(
                            f"""
                            iOS tests configured with a test command that uses the "{{project}}" or
                            "{{package}}" placeholder. iOS tests cannot use placeholders, because the
                            source directory is not visible on the simulator.

                            In addition, iOS tests must run as a Python module, so the test command
                            must begin with 'python -m'.

                            Test command: {build_options.test_command!r}
                            """
                        )
                        raise errors.FatalError(msg)

                    test_command_list = shlex.split(build_options.test_command)
                    try:
                        for test_command_parts in split_command(test_command_list):
                            match test_command_parts:
                                case ["python", "-m", *rest]:
                                    final_command = rest
                                case ["pytest", *rest]:
                                    # pytest works exactly the same as a module, so we
                                    # can just run it as a module.
                                    msg = unwrap_preserving_paragraphs(f"""
                                        iOS tests configured with a test command which doesn't start
                                        with 'python -m'. iOS tests must execute python modules - other
                                        entrypoints are not supported.

                                        cibuildwheel will try to execute it as if it started with
                                        'python -m'. If this works, all you need to do is add that to
                                        your test command.

                                        Test command: {build_options.test_command!r}
                                    """)
                                    log.warning(msg)
                                    final_command = ["pytest", *rest]
                                case _:
                                    msg = unwrap_preserving_paragraphs(
                                        f"""
                                            iOS tests configured with a test command which doesn't start
                                            with 'python -m'. iOS tests must execute python modules - other
                                            entrypoints are not supported.

                                            Test command: {build_options.test_command!r}
                                        """
                                    )
                                    raise errors.FatalError(msg)

                            test_runtime_args = build_options.test_runtime.args

                            # 2025-10: The GitHub Actions macos-15 runner has a known issue where
                            # the default simulator won't start due to a disk performance issue;
                            # see https://github.com/actions/runner-images/issues/12777 for details.
                            # In the meantime, if it looks like we're running on a GitHub Actions
                            # macos-15 runner, use a simulator that is known to work, unless the
                            # user explicitly specifies a simulator.
                            os_version, _, arch = platform.mac_ver()
                            if (
                                "GITHUB_ACTIONS" in os.environ
                                and os_version.startswith("15.")
                                and arch == "arm64"
                                and not any(
                                    arg.startswith("--simulator") for arg in test_runtime_args
                                )
                            ):
                                test_runtime_args = [
                                    "--simulator",
                                    "iPhone 16e,OS=18.5",
                                    *test_runtime_args,
                                ]

                            call(
                                "python",
                                testbed_path,
                                "run",
                                *(["--verbose"] if build_options.build_verbosity > 0 else []),
                                *test_runtime_args,
                                "--",
                                *final_command,
                                env=test_env,
                            )
                    except subprocess.CalledProcessError:
                        # catches the first test command failure in the loop,
                        # implementing short-circuiting
                        log.step_end(success=False)
                        log.error(f"Test suite failed on {config.identifier}")
                        sys.exit(1)

                    log.step_end()

            # We're all done here; move it to output (overwrite existing)
            output_wheel: Path | None = None
            if compatible_wheel is None:
                output_wheel = build_options.output_dir.joinpath(repaired_wheel.name)
                moved_wheel = move_file(repaired_wheel, output_wheel)
                if moved_wheel != output_wheel.resolve():
                    log.warning(
                        f"{repaired_wheel} was moved to {moved_wheel} instead of {output_wheel}"
                    )
                built_wheels.append(output_wheel)

            # Clean up
            shutil.rmtree(identifier_tmp_dir)

            log.build_end(output_wheel)
    except subprocess.CalledProcessError as error:
        msg = f"Command {error.cmd} failed with code {error.returncode}. {error.stdout or ''}"
        raise errors.FatalError(msg) from error



import jinja2

from .base import TestProject
from .c import SPAM_C_TEMPLATE

PYPROJECT_TOML_TEMPLATE = r"""
[build-system]
requires = ["meson-python"]
build-backend = "mesonpy"

[project]
name = "spam"
version = "0.1.0"

[tool.cibuildwheel.windows]
config-settings = { "setup-args" = "--vsenv" }
# building win32 wheels on a 64-bit CI machine with meson-python
# requires a few extra steps outside cibuildwheel. See
# https://github.com/pypa/cibuildwheel/pull/2718
archs = ["auto64"]
"""

MESON_BUILD_TEMPLATE = r"""
project('spam', 'c',
  version: '0.1.0',
  default_options: ['warning_level=2'],
)

py = import('python').find_installation(pure: false)

py.extension_module('spam',
  'spam.c',
  install: true,
)
"""


def new_meson_project(
    *,
    spam_c_top_level_add: str = "",
    spam_c_function_add: str = "",
) -> TestProject:
    project = TestProject()

    project.files.update(
        {
            "spam.c": jinja2.Template(SPAM_C_TEMPLATE),
            "pyproject.toml": jinja2.Template(PYPROJECT_TOML_TEMPLATE),
            "meson.build": jinja2.Template(MESON_BUILD_TEMPLATE),
        }
    )

    project.template_context.update(
        {
            "spam_c_top_level_add": spam_c_top_level_add,
            "spam_c_function_add": spam_c_function_add,
        }
    )

    return project



import platform as platform_module
import re
import shutil
import subprocess
import sys
import typing
from collections.abc import Set
from enum import StrEnum, auto
from typing import Final, Literal, Self

from cibuildwheel import errors
from cibuildwheel.typing import PlatformName

PRETTY_NAMES: Final[dict[PlatformName, str]] = {
    "linux": "Linux",
    "macos": "macOS",
    "windows": "Windows",
    "pyodide": "Pyodide",
    "android": "Android",
    "ios": "iOS",
}

ARCH_SYNONYMS: Final[list[dict[PlatformName, str | None]]] = [
    {"linux": "x86_64", "macos": "x86_64", "windows": "AMD64", "android": "x86_64"},
    {"linux": "i686", "macos": None, "windows": "x86"},
    {"linux": "aarch64", "macos": "arm64", "windows": "ARM64", "android": "arm64_v8a"},
]


def arch_synonym(arch: str, from_platform: PlatformName, to_platform: PlatformName) -> str | None:
    for arch_synonym_ in ARCH_SYNONYMS:
        if arch == arch_synonym_.get(from_platform):
            return arch_synonym_.get(to_platform, arch)

    return arch


def _check_aarch32_el0() -> bool:
    """Check if running armv7l natively on aarch64 is supported"""
    if not sys.platform.startswith("linux"):
        return False
    if platform_module.machine() != "aarch64":
        return False
    executable = shutil.which("linux32")
    if executable is None:
        return False
    check = subprocess.run([executable, "uname", "-m"], check=False, capture_output=True, text=True)
    return check.returncode == 0 and check.stdout.startswith("armv")


@typing.final
class Architecture(StrEnum):
    # mac/linux/android archs
    x86_64 = auto()

    # linux archs
    i686 = auto()
    aarch64 = auto()
    ppc64le = auto()
    s390x = auto()
    armv7l = auto()
    riscv64 = auto()

    # mac archs
    universal2 = auto()
    arm64 = auto()

    # windows archs
    x86 = auto()
    AMD64 = "AMD64"
    ARM64 = "ARM64"

    # WebAssembly
    wasm32 = auto()

    # android archs
    arm64_v8a = auto()

    # iOS "multiarch" architectures that include both
    # the CPU architecture and the ABI.
    arm64_iphoneos = auto()
    arm64_iphonesimulator = auto()
    x86_64_iphonesimulator = auto()

    @classmethod
    def parse_config(cls, config: str, platform: PlatformName) -> set[Self]:
        result = set()
        for arch_str in re.split(r"[\s,]+", config):
            match arch_str:
                case "auto":
                    result |= cls.auto_archs(platform=platform)
                case "native":
                    if native_arch := cls.native_arch(platform=platform):
                        result.add(native_arch)
                case "all":
                    result |= cls.all_archs(platform=platform)
                case "auto64":
                    result |= cls.bitness_archs(platform=platform, bitness="64")
                case "auto32":
                    result |= cls.bitness_archs(platform=platform, bitness="32")
                case _:
                    try:
                        result.add(cls(arch_str))
                    except ValueError as e:
                        msg = f"Invalid architecture '{arch_str}'"
                        raise errors.ConfigurationError(msg) from e
        return result

    @classmethod
    def native_arch(cls, platform: PlatformName) -> Self | None:
        native_machine = platform_module.machine()
        native_architecture = cls(native_machine)

        # Cross-platform support. Used for --print-build-identifiers or docker builds.
        host_platform: PlatformName = (
            "windows"
            if sys.platform.startswith("win")
            else ("macos" if sys.platform.startswith("darwin") else "linux")
        )

        if platform == "pyodide":
            return cls.wasm32
        elif platform == "ios":
            # Can only build for iOS on macOS. The "native" architecture is the
            # simulator for the macOS native platform.
            if host_platform == "macos":
                if native_architecture == cls.x86_64:
                    return cls.x86_64_iphonesimulator
                else:
                    return cls.arm64_iphonesimulator
            else:
                return None

        # we might need to rename the native arch to the machine we're running
        # on, as the same arch can have different names on different platforms
        if host_platform != platform:
            synonym = arch_synonym(native_machine, host_platform, platform)
            if synonym is None:
                # can't build anything on this platform
                return None

            native_architecture = cls(synonym)

        return native_architecture

    @classmethod
    def auto_archs(cls, platform: PlatformName) -> set[Self]:
        native_arch = cls.native_arch(platform)
        if native_arch is None:
            return set()  # can't build anything on this platform
        result = {native_arch}

        match platform:
            case "windows" if cls.AMD64 in result:
                result.add(cls.x86)
            case "ios" if native_arch == cls.arm64_iphonesimulator:
                # Also build the device wheel if we're on ARM64.
                result.add(cls.arm64_iphoneos)

        return result

    @classmethod
    def all_archs(cls, platform: PlatformName) -> set[Self]:
        all_archs_map = {
            "linux": {
                cls.x86_64,
                cls.i686,
                cls.aarch64,
                cls.ppc64le,
                cls.s390x,
                cls.armv7l,
                cls.riscv64,
            },
            "macos": {cls.x86_64, cls.arm64, cls.universal2},
            "windows": {cls.x86, cls.AMD64, cls.ARM64},
            "pyodide": {cls.wasm32},
            "android": {cls.x86_64, cls.arm64_v8a},
            "ios": {
                cls.x86_64_iphonesimulator,
                cls.arm64_iphonesimulator,
                cls.arm64_iphoneos,
            },
        }
        return all_archs_map[platform]

    @classmethod
    def bitness_archs(cls, platform: PlatformName, bitness: Literal["64", "32"]) -> set[Self]:
        # This map maps 64-bit architectures to their 32-bit equivalents.
        archs_map = {
            cls.x86_64: cls.i686,
            cls.AMD64: cls.x86,
            cls.aarch64: cls.armv7l,
        }
        native_arch = cls.native_arch(platform)

        if native_arch is None:
            return set()  # can't build anything on this platform

        if native_arch == cls.wasm32:
            return {native_arch} if bitness == "32" else set()

        match bitness:
            case "64":
                return {native_arch} if native_arch not in archs_map.values() else set()
            case "32":
                if native_arch in archs_map.values():
                    return {native_arch}
                elif native_arch in archs_map and platform in {"linux", "windows"}:
                    if native_arch == cls.aarch64 and not _check_aarch32_el0():
                        # If we're on aarch64, skip if we cannot build armv7l wheels.
                        return set()
                    return {archs_map[native_arch]}
                else:
                    return set()
            case _:
                typing.assert_never(bitness)


def allowed_architectures_check(
    platform: PlatformName,
    architectures: Set[Architecture],
) -> None:
    allowed_architectures = Architecture.all_archs(platform)

    msg = f"{PRETTY_NAMES[platform]} only supports {sorted(allowed_architectures)} at the moment."

    if platform != "linux":
        msg += " If you want to set emulation architectures on Linux, use CIBW_ARCHS_LINUX instead."

    if not architectures <= allowed_architectures:
        msg = f"Invalid archs option {architectures}. " + msg
        raise ValueError(msg)

    if not architectures:
        msg = "Empty archs option set. " + msg
        raise ValueError(msg)



#!/usr/bin/env python3


import difflib
import logging
import operator
import re
import tomllib
from collections.abc import Mapping, MutableMapping
from pathlib import Path
from typing import Any, Final, Literal, TypedDict
from xml.etree import ElementTree as ET

import click
import requests
import rich
from packaging.specifiers import Specifier
from packaging.version import Version
from rich.logging import RichHandler
from rich.syntax import Syntax

from cibuildwheel.extra import dump_python_configurations, get_pyodide_xbuildenv_info
from cibuildwheel.platforms.android import android_triplet

log = logging.getLogger("cibw")

# Looking up the dir instead of using utils.resources_dir
# since we want to write to it.
DIR: Final[Path] = Path(__file__).parent.parent.resolve()
RESOURCES_DIR: Final[Path] = DIR / "cibuildwheel/resources"


ArchStr = Literal["32", "64", "ARM64"]


class Config(TypedDict):
    identifier: str
    version: str


class ConfigUrl(Config):
    url: str


class ConfigPyodide(Config):
    default_pyodide_version: str
    node_version: str


# The following set of "Versions" classes allow the initial call to the APIs to
# be cached and reused in the `update_version_*` methods.


class WindowsVersions:
    def __init__(self, arch_str: ArchStr, free_threaded: bool) -> None:
        response = requests.get("https://api.nuget.org/v3/index.json")
        response.raise_for_status()
        api_info = response.json()

        for resource in api_info["resources"]:
            if resource["@type"] == "PackageBaseAddress/3.0.0":
                endpoint = resource["@id"]

        ARCH_DICT = {"32": "win32", "64": "win_amd64", "ARM64": "win_arm64"}
        PACKAGE_DICT = {"32": "pythonx86", "64": "python", "ARM64": "pythonarm64"}

        self.arch_str = arch_str
        self.arch = ARCH_DICT[arch_str]
        self.free_threaded = free_threaded

        package = PACKAGE_DICT[arch_str]
        if free_threaded:
            package = f"{package}-freethreaded"

        response = requests.get(f"{endpoint}{package}/index.json")
        response.raise_for_status()
        cp_info = response.json()

        self.version_dict = {Version(v): v for v in cp_info["versions"]}

    def update_version_windows(self, spec: Specifier) -> Config | None:
        # Specifier.filter selects all non pre-releases that match the spec,
        # unless there are only pre-releases, then it selects pre-releases
        # instead (like pip)
        unsorted_versions = spec.filter(self.version_dict)
        versions = sorted(unsorted_versions, reverse=True)

        log.debug("Windows %s %s has %s", self.arch, spec, ", ".join(str(v) for v in versions))

        if not versions:
            return None

        flags = "t" if self.free_threaded else ""
        version = versions[0]
        identifier = f"cp{version.major}{version.minor}{flags}-{self.arch}"
        return Config(
            identifier=identifier,
            version=self.version_dict[version],
        )


class GraalPyVersions:
    def __init__(self) -> None:
        response = requests.get("https://api.github.com/repos/oracle/graalpython/releases")
        response.raise_for_status()

        releases = response.json()
        gp_version_re = re.compile(r"-(\d+\.\d+\.\d+)$")
        cp_version_re = re.compile(r"Python (\d+\.\d+(?:\.\d+)?)")
        for release in releases:
            m = gp_version_re.search(release["tag_name"])
            if m:
                release["graalpy_version"] = Version(m.group(1))
            m = cp_version_re.search(release["body"])
            if m:
                release["python_version"] = Version(m.group(1))

        self.releases = [r for r in releases if "graalpy_version" in r and "python_version" in r]

    def update_version(self, identifier: str, spec: Specifier) -> ConfigUrl | None:
        if "x86_64" in identifier or "amd64" in identifier:
            arch = "x86_64"
        elif "arm64" in identifier or "aarch64" in identifier:
            arch = "aarch64"
        else:
            msg = f"{identifier} not supported yet on GraalPy"
            raise RuntimeError(msg)

        gpspec_str = identifier.split("-", maxsplit=1)[0].split("_")[1]
        if "." not in gpspec_str and len(gpspec_str) == 3:
            gpspec_str = gpspec_str[:2] + "." + gpspec_str[-1]
        gpspec = Specifier(f"=={gpspec_str}.*")

        releases_tmp = (r for r in self.releases if spec.contains(r["python_version"]))
        releases_tmp = (r for r in releases_tmp if gpspec.contains(r["graalpy_version"]))
        releases = sorted(releases_tmp, key=lambda r: r["graalpy_version"])

        if not releases:
            msg = f"GraalPy {arch} not found for {spec}!"
            raise RuntimeError(msg)

        release = releases[-1]
        version = release["python_version"]
        gpversion = release["graalpy_version"]

        if "macosx" in identifier:
            arch = "x86_64" if "x86_64" in identifier else "arm64"
            platform = "macos"
        elif "win" in identifier:
            arch = "aarch64" if "arm64" in identifier else "x86_64"
            platform = "windows"
        else:
            msg = "GraalPy provides downloads for macOS and Windows and is included for manylinux"
            raise RuntimeError(msg)

        arch = "amd64" if arch == "x86_64" else "aarch64"
        ext = "zip" if "win" in identifier else "tar.gz"
        urls = [
            rf["browser_download_url"]
            for rf in release["assets"]
            if rf["name"].endswith(f"{platform}-{arch}.{ext}")
            and rf["name"].startswith(f"graalpy-{gpversion.major}")
        ]
        if urls:
            (url,) = urls
            return ConfigUrl(
                identifier=identifier,
                version=f"{version.major}.{version.minor}",
                url=url,
            )
        return None


class PyPyVersions:
    def __init__(self, arch_str: ArchStr):
        response = requests.get("https://downloads.python.org/pypy/versions.json")
        response.raise_for_status()

        releases = [r for r in response.json() if r["pypy_version"] != "nightly"]
        for release in releases:
            release["pypy_version"] = Version(release["pypy_version"])
            release["python_version"] = Version(release["python_version"])

        self.releases = [
            r
            for r in releases
            if not r["pypy_version"].is_prerelease and not r["pypy_version"].is_devrelease
        ]
        self.arch = arch_str

    def get_arch_file(self, release: Mapping[str, Any]) -> str:
        urls: list[str] = [
            rf["download_url"]
            for rf in release["files"]
            if "" in rf["platform"] == f"win{self.arch}"
        ]
        return urls[0] if urls else ""

    def update_version_windows(self, spec: Specifier) -> ConfigUrl:
        releases = [r for r in self.releases if spec.contains(r["python_version"])]
        releases = sorted(releases, key=operator.itemgetter("pypy_version"))
        releases = [r for r in releases if self.get_arch_file(r)]

        if not releases:
            msg = f"PyPy Win {self.arch} not found for {spec}! {self.releases}"
            raise RuntimeError(msg)

        version_arch = "win32" if self.arch == "32" else "win_amd64"

        release = releases[-1]
        version = release["python_version"]
        identifier = f"pp{version.major}{version.minor}-{version_arch}"
        url = self.get_arch_file(release)

        return ConfigUrl(
            identifier=identifier,
            version=f"{version.major}.{version.minor}",
            url=url,
        )

    def update_version_macos(self, spec: Specifier) -> ConfigUrl:
        if self.arch not in {"64", "ARM64"}:
            msg = f"'{self.arch}' arch not supported yet on macOS"
            raise RuntimeError(msg)

        releases = [r for r in self.releases if spec.contains(r["python_version"])]
        releases = sorted(releases, key=operator.itemgetter("pypy_version"))

        if not releases:
            msg = f"PyPy macOS {self.arch} not found for {spec}!"
            raise RuntimeError(msg)

        release = releases[-1]
        version = release["python_version"]
        arch = "x86_64" if self.arch == "64" else self.arch.lower()
        identifier = f"pp{version.major}{version.minor}-macosx_{arch}"

        arch = "x64" if self.arch == "64" else self.arch.lower()
        (url,) = (
            rf["download_url"]
            for rf in release["files"]
            if "" in rf["platform"] == "darwin" and rf["arch"] == arch
        )

        return ConfigUrl(
            identifier=identifier,
            version=f"{version.major}.{version.minor}",
            url=url,
        )


class CPythonVersions:
    def __init__(self) -> None:
        response = requests.get(
            "https://www.python.org/api/v2/downloads/release/?is_published=true"
        )
        response.raise_for_status()

        releases_info = response.json()

        self.versions_dict: dict[Version, int] = {}
        for release in releases_info:
            # Skip the pymanager releases
            if not release["slug"].startswith("python"):
                continue

            # Removing the prefix
            version = Version(release["name"].removeprefix("Python "))
            self.versions_dict[version] = release["resource_uri"]

        files_response = requests.get("https://www.python.org/api/v2/downloads/release_file/")
        files_response.raise_for_status()
        self.files_info = files_response.json()

    def update_version(self, identifier: str, spec: Specifier, file_ident: str) -> ConfigUrl | None:
        # see note above on Specifier.filter
        unsorted_versions = spec.filter(self.versions_dict)
        sorted_versions = sorted(unsorted_versions, reverse=True)

        for new_version in sorted_versions:
            # Find the first patch version that contains the requested file
            uri = self.versions_dict[new_version]
            files = [rf for rf in self.files_info if rf["release"] == uri]

            urls = [rf["url"] for rf in files if file_ident in rf["url"]]
            if urls:
                return ConfigUrl(
                    identifier=identifier,
                    version=f"{new_version.major}.{new_version.minor}",
                    url=urls[0],
                )

        return None

    def update_version_macos(
        self, identifier: str, version: Version, spec: Specifier
    ) -> ConfigUrl | None:
        macver = "x10.9" if version <= Version("3.8.9999") else "11"
        return self.update_version(identifier, spec, f"macos{macver}.pkg")

    def update_version_android(self, identifier: str, spec: Specifier) -> ConfigUrl | None:
        return self.update_version(identifier, spec, android_triplet(identifier))


class MavenVersions:
    MAVEN_URL = "https://repo.maven.apache.org/maven2/com/chaquo/python/python"

    def __init__(self) -> None:
        response = requests.get(f"{self.MAVEN_URL}/maven-metadata.xml")
        response.raise_for_status()
        root = ET.fromstring(response.text)

        self.versions: list[Version] = []
        for version_elem in root.findall("./versioning/versions/version"):
            version_str = version_elem.text
            assert isinstance(version_str, str), version_str
            self.versions.append(Version(version_str))

    def update_version_android(self, identifier: str, spec: Specifier) -> ConfigUrl | None:
        sorted_versions = sorted(spec.filter(self.versions), reverse=True)

        # Return a config using the highest version for the given specifier.
        if sorted_versions:
            max_version = sorted_versions[0]
            triplet = android_triplet(identifier)
            return ConfigUrl(
                identifier=identifier,
                version=f"{max_version.major}.{max_version.minor}",
                url=f"{self.MAVEN_URL}/{max_version}/python-{max_version}-{triplet}.tar.gz",
            )
        else:
            return None


class CPythonIOSVersions:
    def __init__(self) -> None:
        response = requests.get(
            "https://api.github.com/repos/beeware/Python-Apple-support/releases",
            headers={
                "Accept": "application/vnd.github+json",
                "X-Github-Api-Version": "2022-11-28",
            },
        )
        response.raise_for_status()

        releases_info = response.json()
        self.versions_dict: dict[Version, dict[int, str]] = {}

        # Each release has a name like "3.13-b4"
        for release in releases_info:
            py_version, build = release["name"].split("-")
            version = Version(py_version)
            self.versions_dict.setdefault(version, {})

            # There are several release assets associated with each release;
            # The name of the asset will be something like
            # "Python-3.11-iOS-support.b4.tar.gz". Store all builds that are
            # "-iOS-support" builds, retaining the download URL.
            for asset in release["assets"]:
                filename, build, _, _ = asset["name"].rsplit(".", 3)
                if filename.endswith("-iOS-support"):
                    self.versions_dict[version][int(build[1:])] = asset["browser_download_url"]

    def update_version_ios(self, identifier: str, version: Version) -> ConfigUrl | None:
        # Return a config using the highest build number for the given version.
        urls = [url for _, url in sorted(self.versions_dict.get(version, {}).items())]
        if urls:
            return ConfigUrl(
                identifier=identifier,
                version=str(version),
                url=urls[-1],
            )

        return None


class PyodideVersions:
    def __init__(self) -> None:
        xbuildenv_info = get_pyodide_xbuildenv_info()
        self.releases = xbuildenv_info["releases"]

    def update_version_pyodide(
        self, identifier: str, version: Version, spec: Specifier, node_version: str
    ) -> ConfigPyodide | None:
        # get releases that match the python version
        releases = [
            r for r in self.releases.values() if spec.contains(Version(r["python_version"]))
        ]
        # sort by version, latest first
        releases.sort(key=lambda r: Version(r["version"]), reverse=True)

        if not releases:
            msg = f"Pyodide not found for {spec}!"
            raise ValueError(msg)

        final_releases = [r for r in releases if not Version(r["version"]).is_prerelease]

        # prefer a final release if available, otherwise use the latest
        # pre-release
        release = final_releases[0] if final_releases else releases[0]

        return ConfigPyodide(
            identifier=identifier,
            version=str(version),
            default_pyodide_version=release["version"],
            node_version=node_version,
        )


# This is a universal interface to all the above Versions classes. Given an
# identifier, it updates a config dict.


class AllVersions:
    def __init__(self) -> None:
        self.windows_32 = WindowsVersions("32", False)
        self.windows_t_32 = WindowsVersions("32", True)
        self.windows_64 = WindowsVersions("64", False)
        self.windows_t_64 = WindowsVersions("64", True)
        self.windows_arm64 = WindowsVersions("ARM64", False)
        self.windows_t_arm64 = WindowsVersions("ARM64", True)
        self.windows_pypy_64 = PyPyVersions("64")

        self.cpython = CPythonVersions()
        self.macos_pypy = PyPyVersions("64")
        self.macos_pypy_arm64 = PyPyVersions("ARM64")

        self.maven = MavenVersions()
        self.ios_cpython = CPythonIOSVersions()

        self.graalpy = GraalPyVersions()

        self.pyodide = PyodideVersions()

    def update_config(self, config: MutableMapping[str, str]) -> None:
        identifier = config["identifier"]
        version = Version(config["version"])
        spec = Specifier(f"=={version.major}.{version.minor}.*")
        log.info("Reading in %r -> %s @ %s", str(identifier), spec, version)
        config_update: Config | None = None

        # We need to use ** in update due to MyPy (probably a bug)
        if "macosx" in identifier:
            if identifier.startswith("cp"):
                config_update = self.cpython.update_version_macos(identifier, version, spec)
            elif identifier.startswith("pp"):
                if "macosx_x86_64" in identifier:
                    config_update = self.macos_pypy.update_version_macos(spec)
                elif "macosx_arm64" in identifier:
                    config_update = self.macos_pypy_arm64.update_version_macos(spec)
            elif identifier.startswith("gp"):
                if "macosx_x86_64" in identifier:
                    return
                config_update = self.graalpy.update_version(identifier, spec)
        elif "t-win32" in identifier and identifier.startswith("cp"):
            config_update = self.windows_t_32.update_version_windows(spec)
        elif "win32" in identifier and identifier.startswith("cp"):
            config_update = self.windows_32.update_version_windows(spec)
        elif "t-win_amd64" in identifier and identifier.startswith("cp"):
            config_update = self.windows_t_64.update_version_windows(spec)
        elif "win_amd64" in identifier:
            if identifier.startswith("cp"):
                config_update = self.windows_64.update_version_windows(spec)
            elif identifier.startswith("pp"):
                config_update = self.windows_pypy_64.update_version_windows(spec)
            elif identifier.startswith("gp"):
                config_update = self.graalpy.update_version(identifier, spec)
        elif "t-win_arm64" in identifier and identifier.startswith("cp"):
            config_update = self.windows_t_arm64.update_version_windows(spec)
        elif "win_arm64" in identifier and identifier.startswith("cp"):
            config_update = self.windows_arm64.update_version_windows(spec)
        elif "android" in identifier:
            # Python 3.13 is released by Chaquopy on Maven Central.
            # Python 3.14 and newer have official releases on python.org.
            versions = self.maven if identifier.startswith("cp313") else self.cpython
            config_update = versions.update_version_android(identifier, spec)
        elif "ios" in identifier:
            config_update = self.ios_cpython.update_version_ios(identifier, version)
        elif "pyodide" in identifier:
            config_update = self.pyodide.update_version_pyodide(
                identifier, version, spec, config["node_version"]
            )

        assert config_update is not None, f"{identifier} not found!"
        if config_update != config:
            log.info("  Updated %s to %s", config, config_update)
            config.clear()
            config.update(**config_update)


@click.command()
@click.option("--force", is_flag=True)
@click.option(
    "--level", default="INFO", type=click.Choice(["WARNING", "INFO", "DEBUG"], case_sensitive=False)
)
def update_pythons(force: bool, level: str) -> None:
    logging.basicConfig(
        level="INFO",
        format="%(message)s",
        datefmt="[%X]",
        handlers=[RichHandler(rich_tracebacks=True, markup=True)],
    )
    log.setLevel(level)

    all_versions = AllVersions()
    toml_file_path = RESOURCES_DIR / "build-platforms.toml"

    original_toml = toml_file_path.read_text()
    with toml_file_path.open("rb") as f:
        configs = tomllib.load(f)

    for platform in ["windows", "macos", "android", "ios", "pyodide"]:
        for config in configs[platform]["python_configurations"]:
            all_versions.update_config(config)

    result_toml = dump_python_configurations(configs)

    rich.print()  # spacer

    if original_toml == result_toml:
        rich.print("[green]Check complete, Python configurations unchanged.")
        return

    rich.print("Python configurations updated.")
    rich.print("Changes:")
    rich.print()

    toml_relpath = toml_file_path.relative_to(DIR).as_posix()
    diff_lines = difflib.unified_diff(
        original_toml.splitlines(keepends=True),
        result_toml.splitlines(keepends=True),
        fromfile=toml_relpath,
        tofile=toml_relpath,
    )
    rich.print(Syntax("".join(diff_lines), "diff", theme="ansi_light"))
    rich.print()

    if force:
        toml_file_path.write_text(result_toml)
        rich.print("[green]TOML file updated.")
    else:
        rich.print("[yellow]File left unchanged. Use --force flag to update.")


if __name__ == "__main__":
    update_pythons()



# This module is copied into the site-packages directory of an Android build environment, and
# activated via a .pth file when we want the environment to simulate Android.

import os
import platform
import re
import sys
import sysconfig
from pathlib import Path
from typing import Any


def initialize() -> None:
    if not (host_triplet := os.environ.get("CIBW_HOST_TRIPLET")):
        return

    # os ######################################################################
    def cross_os_uname() -> os.uname_result:
        return os.uname_result(
            (
                "Linux",
                "localhost",
                # The Linux kernel version and release are unlikely to be significant, but return
                # realistic values anyway (from an API level 24 emulator).
                "3.18.91+",
                "#1 SMP PREEMPT Tue Jan 9 20:35:43 UTC 2018",
                host_triplet.split("-")[0],
            )
        )

    os.name = "posix"
    os.uname = cross_os_uname

    # platform ################################################################
    #
    # We can't determine the user-visible Android version number from the API level, so return a
    # string which will work fine for display, but will fail to parse as a version number.
    def cross_android_ver(*args: Any, **kwargs: Any) -> platform.AndroidVer:
        return platform.AndroidVer(
            release=f"API level {cross_getandroidapilevel()}",
            api_level=cross_getandroidapilevel(),
            manufacturer="Google",
            model="sdk_gphone64",
            device="emu64",
            is_emulator=True,
        )

    # platform.uname, platform.system etc. are all implemented in terms of platform.android_ver.
    platform.android_ver = cross_android_ver

    # sys #####################################################################
    def cross_getandroidapilevel() -> int:
        api_level = sysconfig.get_config_var("ANDROID_API_LEVEL")
        assert isinstance(api_level, int)
        return api_level

    # Some packages may recognize sys.cross_compiling from the crossenv tool.
    sys.cross_compiling = True  # type: ignore[attr-defined]
    sys.getandroidapilevel = cross_getandroidapilevel  # type: ignore[attr-defined]
    sys.implementation._multiarch = host_triplet  # type: ignore[attr-defined]
    sys.platform = "android"

    # Determine the abiflags from the sysconfigdata filename.
    sysconfigdata_path = next(Path(__file__).parent.glob("_sysconfigdata_*.py"))
    abiflags_match = re.match(r"_sysconfigdata_(.*?)_", sysconfigdata_path.name)
    assert abiflags_match is not None
    sys.abiflags = abiflags_match[1]

    # sysconfig ###############################################################
    #
    # Load the sysconfigdata file, generating its name from sys.abiflags,
    # sys.platform, and sys.implementation._multiarch.
    sysconfig._init_config_vars()  # type: ignore[attr-defined]

    # We don't change the actual sys.base_prefix and base_exec_prefix, because that
    # could have unpredictable effects. Instead, we change the sysconfig variables
    # used by sysconfig.get_paths().
    vars = sysconfig.get_config_vars()
    try:
        host_prefix = vars["host_prefix"]  # This variable was added in Python 3.14.
    except KeyError:
        host_prefix = vars["exec_prefix"]
    vars["installed_base"] = vars["installed_platbase"] = host_prefix

    # sysconfig.get_platform, which determines the wheel tag, is implemented in terms of
    # sys.platform, sysconfig.get_config_var("ANDROID_API_LEVEL") (see localized_vars in
    # android.py), and os.uname.



from pathlib import Path
from typing import Any, Self

import jinja2

FilesDict = dict[str, str | jinja2.Template]
TemplateContext = dict[str, Any]


class TestProject:
    """
    An object that represents a project that can be built by cibuildwheel.
    Can be manipulated in tests by changing `files` and `template_context`.

    Write out to the filesystem using `generate`.
    """

    __test__ = False  # Have pytest ignore this class on `from .test_projects import TestProject`

    files: FilesDict
    template_context: TemplateContext

    def __init__(self) -> None:
        self.files = {}
        self.template_context = {}

    def generate(self, path: Path) -> None:
        for filename, content in self.files.items():
            file_path = path / filename
            file_path.parent.mkdir(parents=True, exist_ok=True)

            with file_path.open("w", encoding="utf8") as f:
                if isinstance(content, jinja2.Template):
                    content = content.render(self.template_context)  # noqa: PLW2901

                f.write(content)

    def copy(self) -> Self:
        other = self.__class__()
        other.files = self.files.copy()
        other.template_context = self.template_context.copy()
        return other



SPAM_C_TEMPLATE = r"""
#include <Python.h>

{{ spam_c_top_level_add }}

static PyObject *
spam_filter(PyObject *self, PyObject *args)
{
    const char *content;
    int sts;

    if (!PyArg_ParseTuple(args, "s", &content))
        return NULL;

    // Spam should not be allowed through the filter.
    sts = strcmp(content, "spam") != 0;

    {{ spam_c_function_add | indent(4) }}

    return PyLong_FromLong(sts);
}

/* Module initialization */
static PyMethodDef module_methods[] = {
    {"filter", (PyCFunction)spam_filter, METH_VARARGS,
     "Execute a shell command."},
    {NULL}  /* Sentinel */
};

PyMODINIT_FUNC PyInit_spam(void)
{
    static struct PyModuleDef moduledef = {
        PyModuleDef_HEAD_INIT, "spam", "Example module", -1, module_methods,
    };
    return PyModule_Create(&moduledef);
}
"""



# this file is copied to the testing cwd, to raise the below error message if
# pytest/unittest is run from there.

import sys
import unittest
from typing import NoReturn


class TestStringMethods(unittest.TestCase):
    def test_fail(self) -> NoReturn:
        if sys.platform in ["android", "ios"]:
            msg = (
                "You tried to run tests from the testbed app's working "
                "directory, without specifying `test-sources`. "
                "On this platform, you must copy your test files to the testbed app by "
                "setting the `test-sources` option in your cibuildwheel "
                "configuration."
            )
        else:
            msg = (
                "cibuildwheel executes tests from a different working directory to "
                "your project. This ensures only your wheel is imported, preventing "
                "Python from accessing files that haven't been packaged into the "
                "wheel. "
                "\n\n"
                "Please specify a path to your tests when invoking pytest "
                "using the {project} placeholder, e.g. `pytest {project}` or "
                "`pytest {project}/tests`. cibuildwheel will replace {project} with "
                "the path to your project. "
                "\n\n"
                "Alternatively, you can specify your test files using the "
                "`test-sources` option, and cibuildwheel will copy them to the "
                "working directory for testing."
            )

        self.fail(msg)



#!/usr/bin/env python3


import subprocess
import sys
from pathlib import Path

try:
    from html2image import Html2Image  # type: ignore[import-not-found]
except ImportError:
    sys.exit(
        """
        html2image not found. Ensure you have Chrome (on Mac/Windows) or
        Chromium (on Linux) installed, and then do:
            pip install html2image
        """
    )


def main() -> None:
    subprocess.run(["mkdocs", "build"], check=True)

    hti = Html2Image(custom_flags=["--force-device-scale-factor=2"])

    html_str = Path("docs/diagram.md").read_text()
    css_tags = f"""
        <style>{Path("site/css/theme.css").read_text()}</style>
        <style>{Path("site/css/theme_extra.css").read_text()}</style>
        <style>{Path("site/extra.css").read_text()}</style>
        <style>
            body {{
                background: white;
            }}
        </style>
    """
    html_str = css_tags + html_str

    [screenshot, *_] = hti.screenshot(
        html_str=html_str,
        size=(830, 405),
    )

    dest_path = Path("docs/data/how-it-works.png")
    dest_path.unlink(missing_ok=True)

    Path(screenshot).rename(dest_path)


if __name__ == "__main__":
    main()



import codecs
import contextlib
import dataclasses
import functools
import hashlib
import io
import os
import re
import sys
import textwrap
import time
from collections.abc import Generator
from pathlib import Path
from typing import IO, TYPE_CHECKING, AnyStr, Final, Literal

import humanize

from cibuildwheel.ci import CIProvider, detect_ci_provider, filter_ansi_codes

if TYPE_CHECKING:
    from cibuildwheel.options import Options

FoldPattern = tuple[str, str]
DEFAULT_FOLD_PATTERN: Final[FoldPattern] = ("{name}", "")
FOLD_PATTERNS: Final[dict[str, FoldPattern]] = {
    "azure": ("##[group]{name}", "##[endgroup]"),
    "travis": ("travis_fold:start:{identifier}\n{name}", "travis_fold:end:{identifier}"),
    "github": ("::group::{name}", "::endgroup::{name}"),
}

PLATFORM_IDENTIFIER_DESCRIPTIONS: Final[dict[str, str]] = {
    "manylinux_x86_64": "manylinux x86_64",
    "manylinux_i686": "manylinux i686",
    "manylinux_aarch64": "manylinux aarch64",
    "manylinux_ppc64le": "manylinux ppc64le",
    "manylinux_s390x": "manylinux s390x",
    "manylinux_armv7l": "manylinux armv7l",
    "manylinux_riscv64": "manylinux riscv64",
    "musllinux_x86_64": "musllinux x86_64",
    "musllinux_i686": "musllinux i686",
    "musllinux_aarch64": "musllinux aarch64",
    "musllinux_ppc64le": "musllinux ppc64le",
    "musllinux_s390x": "musllinux s390x",
    "musllinux_armv7l": "musllinux armv7l",
    "musllinux_riscv64": "musllinux riscv64",
    "win32": "Windows 32bit",
    "win_amd64": "Windows 64bit",
    "win_arm64": "Windows on ARM 64bit",
    "macosx_x86_64": "macOS x86_64",
    "macosx_universal2": "macOS Universal 2 - x86_64 and arm64",
    "macosx_arm64": "macOS arm64 - Apple Silicon",
    "pyodide_wasm32": "Pyodide",
    "android_arm64_v8a": "Android arm64_v8a",
    "android_x86_64": "Android x86_64",
    "ios_arm64_iphoneos": "iOS Device (ARM64)",
    "ios_arm64_iphonesimulator": "iOS Simulator (ARM64)",
    "ios_x86_64_iphonesimulator": "iOS Simulator (x86_64)",
}


class Colors:
    def __init__(self, *, enabled: bool) -> None:
        self.red = "\033[31m" if enabled else ""
        self.green = "\033[32m" if enabled else ""
        self.yellow = "\033[33m" if enabled else ""
        self.blue = "\033[34m" if enabled else ""
        self.cyan = "\033[36m" if enabled else ""
        self.bright_red = "\033[91m" if enabled else ""
        self.bright_green = "\033[92m" if enabled else ""
        self.white = "\033[37m\033[97m" if enabled else ""
        self.gray = "\033[38;5;244m" if enabled else ""

        self.bg_grey = "\033[48;5;235m" if enabled else ""

        self.bold = "\033[1m" if enabled else ""
        self.faint = "\033[2m" if enabled else ""

        self.end = "\033[0m" if enabled else ""


class Symbols:
    def __init__(self, *, unicode: bool) -> None:
        self.done = "✓" if unicode else "done"
        self.error = "✕" if unicode else "failed"


@dataclasses.dataclass(kw_only=True, frozen=True)
class BuildInfo:
    identifier: str
    filename: Path | None
    duration: float

    @functools.cached_property
    def size(self) -> str | None:
        if self.filename is None:
            return None
        return humanize.naturalsize(self.filename.stat().st_size)

    @functools.cached_property
    def sha256(self) -> str | None:
        if self.filename is None:
            return None
        with self.filename.open("rb") as f:
            digest = hashlib.file_digest(f, "sha256")
        return digest.hexdigest()

    def __str__(self) -> str:
        duration = humanize.naturaldelta(self.duration)
        if self.filename:
            return f"{self.identifier}: {self.filename.name} {self.size} in {duration}, SHA256={self.sha256}"
        return f"{self.identifier}: {duration} (test only)"


class Logger:
    fold_mode: Literal["azure", "github", "travis", "disabled"]
    colors_enabled: bool
    unicode_enabled: bool
    active_build_identifier: str | None = None
    build_start_time: float | None = None
    step_start_time: float | None = None
    active_fold_group_name: str | None = None
    summary: list[BuildInfo]

    def __init__(self) -> None:
        if sys.platform == "win32" and hasattr(sys.stdout, "reconfigure"):
            # the encoding on Windows can be a 1-byte charmap, but all CIs
            # support utf8, so we hardcode that
            sys.stdout.reconfigure(encoding="utf8")

        self.unicode_enabled = file_supports_unicode(sys.stdout)

        ci_provider = detect_ci_provider()

        match ci_provider:
            case CIProvider.azure_pipelines:
                self.fold_mode = "azure"
                self.colors_enabled = True

            case CIProvider.github_actions:
                self.fold_mode = "github"
                self.colors_enabled = True

            case CIProvider.travis_ci:
                self.fold_mode = "travis"
                self.colors_enabled = True

            case CIProvider.appveyor:
                self.fold_mode = "disabled"
                self.colors_enabled = True

            case _:
                self.fold_mode = "disabled"
                self.colors_enabled = file_supports_color(sys.stdout)

        self.summary = []

    def build_start(self, identifier: str) -> None:
        self.step_end()
        c = self.colors
        description = build_description_from_identifier(identifier)
        print()
        print(f"{c.bold}{c.blue}Building {identifier} wheel{c.end}")
        print(f"{description}")
        print()

        self.build_start_time = time.time()
        self.active_build_identifier = identifier

    def build_end(self, filename: Path | None) -> None:
        assert self.build_start_time is not None
        assert self.active_build_identifier is not None
        self.step_end()

        c = self.colors
        s = self.symbols
        duration = time.time() - self.build_start_time
        duration_str = humanize.naturaldelta(duration, minimum_unit="milliseconds")

        print()
        print(f"{c.green}{s.done} {c.end}{self.active_build_identifier} finished in {duration_str}")
        self.summary.append(
            BuildInfo(identifier=self.active_build_identifier, filename=filename, duration=duration)
        )

        self.build_start_time = None
        self.active_build_identifier = None

    def step(self, step_description: str) -> None:
        self.step_end()
        self.step_start_time = time.time()
        self._start_fold_group(step_description)

    def step_end(self, success: bool = True) -> None:
        if self.step_start_time is not None:
            self._end_fold_group()
            c = self.colors
            s = self.symbols
            duration = time.time() - self.step_start_time

            if success:
                print(f"{c.green}{s.done} {c.end}{duration:.2f}s".rjust(78))
            else:
                print(f"{c.red}{s.error} {c.end}{duration:.2f}s".rjust(78))

            self.step_start_time = None

    def step_end_with_error(self, error: BaseException | str) -> None:
        self.step_end(success=False)
        self.error(error)

    def quiet(self, message: str) -> None:
        c = self.colors
        print(f"{c.gray}{message}{c.end}", file=sys.stderr)

    def notice(self, message: str) -> None:
        if self.fold_mode == "github":
            print(f"::notice::cibuildwheel: {message}\n", file=sys.stderr)
        else:
            c = self.colors
            print(f"cibuildwheel: {c.bold}note{c.end}: {message}\n", file=sys.stderr)

    def warning(self, message: str) -> None:
        if self.fold_mode == "github":
            print(f"::warning::cibuildwheel: {message}\n", file=sys.stderr)
        else:
            c = self.colors
            print(f"cibuildwheel: {c.yellow}warning{c.end}: {message}\n", file=sys.stderr)

    def error(self, error: BaseException | str) -> None:
        if self.fold_mode == "github":
            print(f"::error::cibuildwheel: {error}\n", file=sys.stderr)
        else:
            c = self.colors
            print(f"cibuildwheel: {c.bright_red}error{c.end}: {error}\n", file=sys.stderr)

    @contextlib.contextmanager
    def print_summary(self, *, options: "Options") -> Generator[None, None, None]:
        start = time.time()
        yield
        duration = time.time() - start
        if summary_path := os.environ.get("GITHUB_STEP_SUMMARY"):
            github_summary = self._github_step_summary(duration=duration, options=options)
            Path(summary_path).write_text(filter_ansi_codes(github_summary), encoding="utf-8")

        n_wheels = len([info for info in self.summary if info.filename])
        s = "s" if n_wheels > 1 else ""
        duration_str = humanize.naturaldelta(duration)
        print()
        self._start_fold_group(f"{n_wheels} wheel{s} produced in {duration_str}")
        for build_info in self.summary:
            print(" ", build_info)
        self._end_fold_group()

        self.summary = []

    @property
    def step_active(self) -> bool:
        return self.step_start_time is not None

    def _start_fold_group(self, name: str) -> None:
        self._end_fold_group()
        self.active_fold_group_name = name
        fold_start_pattern = FOLD_PATTERNS.get(self.fold_mode, DEFAULT_FOLD_PATTERN)[0]
        identifier = self._fold_group_identifier(name)

        print(fold_start_pattern.format(name=self.active_fold_group_name, identifier=identifier))
        print()
        sys.stdout.flush()

    def _end_fold_group(self) -> None:
        if self.active_fold_group_name:
            fold_start_pattern = FOLD_PATTERNS.get(self.fold_mode, DEFAULT_FOLD_PATTERN)[1]
            identifier = self._fold_group_identifier(self.active_fold_group_name)
            print(
                fold_start_pattern.format(name=self.active_fold_group_name, identifier=identifier)
            )
            sys.stdout.flush()
            self.active_fold_group_name = None

    @staticmethod
    def _fold_group_identifier(name: str) -> str:
        """
        Travis doesn't like fold groups identifiers that have spaces in. This
        method converts them to ascii identifiers
        """
        # whitespace to underscores
        identifier = re.sub(r"\s+", "_", name)
        # remove non-alphanum
        identifier = re.sub(r"[^A-Za-z\d_]+", "", identifier)
        # trim underscores
        identifier = identifier.strip("_")
        # lowercase, shorten
        return identifier.lower()[:20]

    def _github_step_summary(self, duration: float, options: "Options") -> str:
        """
        Returns the GitHub step summary, in markdown format.
        """
        out = io.StringIO()
        options_summary = options.summary(
            identifiers=[bi.identifier for bi in self.summary], skip_unset=True
        )
        out.write(
            textwrap.dedent("""\
                ### 🎡 cibuildwheel

                <details>
                <summary>
                Build options
                </summary>

                ```yaml
                {options_summary}
                ```

                </details>

            """).format(options_summary=options_summary)
        )
        n_wheels = len([b for b in self.summary if b.filename])
        wheel_rows = "\n".join(
            "<tr>"
            f"<td nowrap>{'<samp>' + b.filename.name + '</samp>' if b.filename else '*Test only*'}</td>"
            f"<td nowrap>{b.size or 'N/A'}</td>"
            f"<td nowrap><samp>{b.identifier}</samp></td>"
            f"<td nowrap>{humanize.naturaldelta(b.duration)}</td>"
            f"<td nowrap><samp>{b.sha256 or 'N/A'}</samp></td>"
            "</tr>"
            for b in self.summary
        )
        out.write(
            textwrap.dedent("""\
                <table>
                <thead>
                <tr>
                <th align="left">Wheel</th>
                <th align="left">Size</th>
                <th align="left">Build identifier</th>
                <th align="left">Time</th>
                <th align="left">SHA256</th>
                </tr>
                </thead>
                <tbody>
                {wheel_rows}
                </tbody>
                </table>
                <div align="right"><sup>{n} wheel{s} created in {duration_str}</sup></div>
            """).format(
                wheel_rows=wheel_rows,
                n=n_wheels,
                duration_str=humanize.naturaldelta(duration),
                s="s" if n_wheels > 1 else "",
            )
        )

        out.write("\n")
        out.write("---")
        out.write("\n")
        return out.getvalue()

    @property
    def colors(self) -> Colors:
        return Colors(enabled=self.colors_enabled)

    @property
    def symbols(self) -> Symbols:
        return Symbols(unicode=self.unicode_enabled)


def build_description_from_identifier(identifier: str) -> str:
    python_identifier, _, platform_identifier = identifier.partition("-")

    build_description = ""

    python_interpreter = python_identifier[0:2]
    version_parts = python_identifier[2:].split("_")
    python_version = version_parts[0]

    if python_interpreter == "cp":
        build_description += "CPython"
    elif python_interpreter == "pp":
        build_description += "PyPy"
    elif python_interpreter == "gp":
        build_description += "GraalPy"
    else:
        msg = f"unknown python {python_interpreter!r}"
        raise Exception(msg)

    build_description += f" {python_version[0]}.{python_version[1:]} "
    if len(version_parts) > 1:
        build_description += f"(ABI {version_parts[1]}) "

    try:
        build_description += PLATFORM_IDENTIFIER_DESCRIPTIONS[platform_identifier]
    except KeyError as e:
        msg = f"unknown platform {platform_identifier!r}"
        raise Exception(msg) from e

    return build_description


def file_supports_color(file_obj: IO[AnyStr]) -> bool:
    """
    Returns True if the running system's terminal supports color.
    """
    plat = sys.platform
    supported_platform = plat != "win32" or "ANSICON" in os.environ

    is_a_tty = file_is_a_tty(file_obj)

    return supported_platform and is_a_tty


def file_is_a_tty(file_obj: IO[AnyStr]) -> bool:
    return hasattr(file_obj, "isatty") and file_obj.isatty()


def file_supports_unicode(file_obj: IO[AnyStr]) -> bool:
    encoding = getattr(file_obj, "encoding", None)
    if not encoding:
        return False

    codec_info = codecs.lookup(encoding)

    return "utf" in codec_info.name


# Global instance of the Logger.
# (there's only one stdout per-process, so a global instance is justified)
log = Logger()



import collections
import configparser
import contextlib
import dataclasses
import difflib
import enum
import functools
import shlex
import textwrap
import tomllib
from collections.abc import Callable, Generator, Iterable, Mapping, Sequence, Set
from pathlib import Path
from typing import Any, Final, Literal, Self, assert_never

from packaging.specifiers import SpecifierSet

from cibuildwheel import errors
from cibuildwheel.architecture import Architecture
from cibuildwheel.environment import EnvironmentParseError, ParsedEnvironment, parse_environment
from cibuildwheel.frontend import BuildFrontendConfig
from cibuildwheel.logger import log
from cibuildwheel.oci_container import OCIContainerEngineConfig
from cibuildwheel.projectfiles import get_requires_python_str, resolve_dependency_groups
from cibuildwheel.selector import BuildSelector, EnableGroup, TestSelector, selector_matches
from cibuildwheel.typing import PLATFORMS, PlatformName
from cibuildwheel.util import resources
from cibuildwheel.util.helpers import format_safe, parse_key_value_string, strtobool, unwrap
from cibuildwheel.util.packaging import DependencyConstraints

MANYLINUX_ARCHS: Final[tuple[str, ...]] = (
    "x86_64",
    "i686",
    "pypy_x86_64",
    "aarch64",
    "ppc64le",
    "s390x",
    "armv7l",
    "riscv64",
    "pypy_aarch64",
    "pypy_i686",
)

MUSLLINUX_ARCHS: Final[tuple[str, ...]] = (
    "x86_64",
    "i686",
    "aarch64",
    "ppc64le",
    "s390x",
    "armv7l",
    "riscv64",
)


@dataclasses.dataclass(kw_only=True)
class CommandLineArguments:
    platform: Literal["auto", "linux", "macos", "windows"] | None
    archs: str | None
    output_dir: Path
    only: str | None
    config_file: str
    package_dir: Path
    print_build_identifiers: bool
    allow_empty: bool
    debug_traceback: bool
    enable: list[str]
    clean_cache: bool

    @classmethod
    def defaults(cls) -> Self:
        return cls(
            platform="auto",
            allow_empty=False,
            archs=None,
            only=None,
            config_file="",
            output_dir=Path("wheelhouse"),
            package_dir=Path(),
            print_build_identifiers=False,
            debug_traceback=False,
            enable=[],
            clean_cache=False,
        )


@dataclasses.dataclass(frozen=True, kw_only=True)
class GlobalOptions:
    package_dir: Path
    output_dir: Path
    build_selector: BuildSelector
    test_selector: TestSelector
    architectures: set[Architecture]
    allow_empty: bool


@dataclasses.dataclass(frozen=True)
class TestRuntimeConfig:
    args: Sequence[str] = ()

    @classmethod
    def from_config_string(cls, config_string: str) -> Self:
        config_dict = parse_key_value_string(config_string, [], ["args"])
        args = config_dict.get("args") or []
        return cls(args=args)

    def options_summary(self) -> str | dict[str, str]:
        return {"args": repr(self.args)}


@dataclasses.dataclass(frozen=True, kw_only=True)
class BuildOptions:
    globals: GlobalOptions
    environment: ParsedEnvironment
    before_all: str
    before_build: str | None
    xbuild_tools: list[str] | None
    repair_command: str
    manylinux_images: dict[str, str] | None
    musllinux_images: dict[str, str] | None
    dependency_constraints: DependencyConstraints
    test_command: str | None
    before_test: str | None
    test_sources: list[str]
    test_requires: list[str]
    test_extras: str
    test_groups: list[str]
    test_environment: ParsedEnvironment
    test_runtime: TestRuntimeConfig
    build_verbosity: int
    build_frontend: BuildFrontendConfig
    config_settings: str
    container_engine: OCIContainerEngineConfig
    pyodide_version: str | None

    @property
    def package_dir(self) -> Path:
        return self.globals.package_dir

    @property
    def output_dir(self) -> Path:
        return self.globals.output_dir

    @property
    def build_selector(self) -> BuildSelector:
        return self.globals.build_selector

    @property
    def test_selector(self) -> TestSelector:
        return self.globals.test_selector

    @property
    def architectures(self) -> set[Architecture]:
        return self.globals.architectures


SettingLeaf = str | int | bool
SettingList = Sequence[SettingLeaf]
SettingTable = Mapping[str, SettingLeaf | SettingList]
SettingValue = SettingTable | SettingList | SettingLeaf


class InheritRule(enum.Enum):
    NONE = enum.auto()
    APPEND = enum.auto()
    PREPEND = enum.auto()


@dataclasses.dataclass(frozen=True)
class Override:
    select_pattern: str
    options: dict[str, SettingValue]
    inherit: dict[str, InheritRule]


MANYLINUX_OPTIONS = {f"manylinux-{build_platform}-image" for build_platform in MANYLINUX_ARCHS}
MUSLLINUX_OPTIONS = {f"musllinux-{build_platform}-image" for build_platform in MUSLLINUX_ARCHS}
DISALLOWED_OPTIONS = {
    "linux": {"dependency-versions"},
    "macos": MANYLINUX_OPTIONS | MUSLLINUX_OPTIONS,
    "windows": MANYLINUX_OPTIONS | MUSLLINUX_OPTIONS,
}


class OptionsReaderError(errors.ConfigurationError):
    pass


class OptionFormat:
    """
    Base class for option format specifiers. These objects describe how values
    can be parsed from rich TOML values and how they're merged together.
    """

    class NotSupported(Exception):
        pass

    def format_list(self, value: SettingList) -> str:  # noqa: ARG002
        raise OptionFormat.NotSupported

    def format_table(self, table: SettingTable) -> str:  # noqa: ARG002
        raise OptionFormat.NotSupported

    def merge_values(self, before: str, after: str) -> str:  # noqa: ARG002
        raise OptionFormat.NotSupported


class ListFormat(OptionFormat):
    """
    A format that joins lists with a separator.
    """

    def __init__(self, sep: str, quote: Callable[[str], str] | None = None) -> None:
        self.sep = sep
        self.quote = quote or (lambda s: s)

    def format_list(self, value: SettingList) -> str:
        return self.sep.join(self.quote(str(v)) for v in value)

    def merge_values(self, before: str, after: str) -> str:
        return f"{before}{self.sep}{after}"


class ShlexTableFormat(OptionFormat):
    """
    The standard table format uses shlex.quote to quote values and shlex.split
    to unquote and split them. When merging values, keys in before are
    replaced by keys in after.
    """

    def __init__(self, sep: str = " ", pair_sep: str = "=", allow_merge: bool = True) -> None:
        self.sep = sep
        self.pair_sep = pair_sep
        self.allow_merge = allow_merge

    def format_table(self, table: SettingTable) -> str:
        assignments: list[tuple[str, str]] = []

        for k, v in table.items():
            if shlex.split(k) != [k]:
                msg = f"Invalid table key: {k}"
                raise OptionsReaderError(msg)

            if isinstance(v, str):
                assignments.append((k, v))
            elif isinstance(v, Sequence):
                assignments.extend((k, str(inner_v)) for inner_v in v)
            else:
                assignments.append((k, str(v)))

        return self.sep.join(f"{k}{self.pair_sep}{shlex.quote(v)}" for k, v in assignments)

    def merge_values(self, before: str, after: str) -> str:
        if not self.allow_merge:
            raise OptionFormat.NotSupported

        before_dict = self.parse_table(before)
        after_dict = self.parse_table(after)

        return self.format_table({**before_dict, **after_dict})

    def parse_table(self, table: str) -> Mapping[str, str | Sequence[str]]:
        assignments: list[tuple[str, str]] = []

        for assignment_str in shlex.split(table):
            key, sep, value = assignment_str.partition(self.pair_sep)

            if not sep:
                msg = f"malformed option with value {assignment_str!r}"
                raise OptionsReaderError(msg)

            assignments.append((key, value))

        result: dict[str, str | list[str]] = {}

        for key, value in assignments:
            if key in result:
                existing_value = result[key]
                if isinstance(existing_value, list):
                    result[key] = [*existing_value, value]
                else:
                    result[key] = [existing_value, value]
            else:
                result[key] = value

        return result


class EnvironmentFormat(OptionFormat):
    """
    The environment format accepts a table of environment variables, where the
    values may contain variables or command substitutions.
    """

    def format_table(self, table: SettingTable) -> str:
        return " ".join(f'{k}="{v}"' for k, v in table.items())

    def merge_values(self, before: str, after: str) -> str:
        return f"{before} {after}"


def _resolve_cascade(
    *pairs: tuple[SettingValue | None, InheritRule],
    ignore_empty: bool = False,
    option_format: OptionFormat | None = None,
) -> str:
    """
    Given a cascade of values with inherit rules, resolve them into a single
    value.

    'None' values mean that the option was not set at that level, and are
    ignored. If `ignore_empty` is True, empty values are ignored too.

    Values start with defaults, followed by more specific rules. If rules are
    NONE, the last non-null value is returned. If a rule is APPEND or PREPEND,
    the value is concatenated with the previous value.

    The following idiom can be used to get the first matching value:

        _resolve_cascade(("value1", Inherit.NONE), ("value2", Inherit.NONE), ...)))
    """
    if not pairs:
        msg = "pairs cannot be empty"
        raise ValueError(msg)

    result: str | None = None

    for value, rule in pairs:
        if value is None:
            continue

        if ignore_empty and not value and value is not False:
            continue

        value_string = _stringify_setting(value, option_format=option_format)

        result = _apply_inherit_rule(result, value_string, rule=rule, option_format=option_format)

    if result is None:
        msg = "a setting should at least have a default value"
        raise ValueError(msg)

    return result


def _apply_inherit_rule(
    before: str | None, after: str, rule: InheritRule, option_format: OptionFormat | None
) -> str:
    if rule == InheritRule.NONE:
        return after

    if not before:
        # if before is None, we can just return after
        # if before is an empty string, we shouldn't add any separator
        return after

    if not after:
        # if after is an empty string, we shouldn't add any separator
        return before

    if not option_format:
        msg = f"Don't know how to merge {before!r} and {after!r} with {rule}"
        raise OptionsReaderError(msg)

    match rule:
        case InheritRule.APPEND:
            return option_format.merge_values(before, after)
        case InheritRule.PREPEND:
            return option_format.merge_values(after, before)
        case _:
            assert_never(rule)


def _stringify_setting(
    setting: SettingValue,
    option_format: OptionFormat | None,
) -> str:
    match setting:
        case {}:
            assert isinstance(setting, Mapping)  # MyPy 1.15 doesn't narrow this for us
            try:
                if option_format is None:
                    raise OptionFormat.NotSupported
                return option_format.format_table(setting)
            except OptionFormat.NotSupported:
                msg = (
                    f"Error converting {setting!r} to a string: this setting doesn't accept a table"
                )
                raise OptionsReaderError(msg) from None
        case bool() | int():
            return str(setting)
        case [*_]:
            try:
                if option_format is None:
                    raise OptionFormat.NotSupported
                return option_format.format_list(setting)
            except OptionFormat.NotSupported:
                msg = (
                    f"Error converting {setting!r} to a string: this setting doesn't accept a list"
                )
                raise OptionsReaderError(msg) from None
        case _:
            assert isinstance(setting, str)  # MyPy 1.15 doesn't narrow this for us
            return setting


class OptionsReader:
    """
    Gets options from the environment, config or defaults, optionally scoped
    by the platform.

    Example:
      >>> # xdoctest: +SKIP
      >>> options_reader = OptionsReader(config_file, platform='macos')
      >>> options_reader.get('cool-color')

      This will return the value of CIBW_COOL_COLOR_MACOS if it exists,
      otherwise the value of CIBW_COOL_COLOR, otherwise
      'tool.cibuildwheel.macos.cool-color' or 'tool.cibuildwheel.cool-color'
      from `config_file`, or from cibuildwheel/resources/defaults.toml. An
      error is thrown if there are any unexpected keys or sections in
      tool.cibuildwheel.
    """

    def __init__(
        self,
        config_file_path: Path | None = None,
        *,
        platform: PlatformName,
        env: Mapping[str, str],
        disallow: Mapping[str, Set[str]] | None = None,
    ) -> None:
        self.platform = platform
        self.env = env
        self.disallow = disallow or {}

        # Open defaults.toml, loading both global and platform sections
        self.default_options, self.default_platform_options = self._load_file(resources.DEFAULTS)

        # Load the project config file
        config_options: dict[str, Any] = {}
        config_platform_options: dict[str, Any] = {}

        if config_file_path is not None:
            config_options, config_platform_options = self._load_file(config_file_path)

        # Validate project config
        for option_name in config_options:
            self._validate_global_option(option_name)

        for option_name in config_platform_options:
            self._validate_platform_option(option_name)

        self.config_options = config_options
        self.config_platform_options = config_platform_options

        self.overrides: list[Override] = []
        self.current_identifier: str | None = None

        config_overrides = self.config_options.get("overrides")

        if config_overrides is not None:
            if not isinstance(config_overrides, list):
                msg = "'tool.cibuildwheel.overrides' must be a list"
                raise OptionsReaderError(msg)

            for config_override in config_overrides:
                select = config_override.pop("select", None)

                if not select:
                    msg = "'select' must be set in an override"
                    raise OptionsReaderError(msg)

                if isinstance(select, list):
                    select = " ".join(select)

                inherit = config_override.pop("inherit", {})
                if not isinstance(inherit, dict) or not all(
                    i in {"none", "append", "prepend"} for i in inherit.values()
                ):
                    msg = "'inherit' must be a dict containing only {'none', 'append', 'prepend'} values"
                    raise OptionsReaderError(msg)

                inherit_enum = {k: InheritRule[v.upper()] for k, v in inherit.items()}

                self.overrides.append(Override(select, config_override, inherit_enum))

    def _validate_global_option(self, name: str) -> None:
        """
        Raises an error if an option with this name is not allowed in the
        [tool.cibuildwheel] section of a config file.
        """
        allowed_option_names = self.default_options.keys() | PLATFORMS | {"overrides"}

        if name not in allowed_option_names:
            msg = f"Option {name!r} not supported in a config file."
            matches = difflib.get_close_matches(name, allowed_option_names, 1, 0.7)
            if matches:
                msg += f" Perhaps you meant {matches[0]!r}?"
            raise OptionsReaderError(msg)

    def _validate_platform_option(self, name: str) -> None:
        """
        Raises an error if an option with this name is not allowed in the
        [tool.cibuildwheel.<current-platform>] section of a config file.
        """
        disallowed_platform_options = self.disallow.get(self.platform, set())
        if name in disallowed_platform_options:
            msg = f"{name!r} is not allowed in {disallowed_platform_options}"
            raise OptionsReaderError(msg)

        allowed_option_names = self.default_options.keys() | self.default_platform_options.keys()

        if name not in allowed_option_names:
            msg = f"Option {name!r} not supported in the {self.platform!r} section"
            matches = difflib.get_close_matches(name, allowed_option_names, 1, 0.7)
            if matches:
                msg += f" Perhaps you meant {matches[0]!r}?"
            raise OptionsReaderError(msg)

    def _load_file(self, filename: Path) -> tuple[dict[str, Any], dict[str, Any]]:
        """
        Load a toml file, returns global and platform as separate dicts.
        """
        with filename.open("rb") as f:
            config = tomllib.load(f)

        global_options = config.get("tool", {}).get("cibuildwheel", {})
        platform_options = global_options.get(self.platform, {})

        return global_options, platform_options

    @property
    def active_config_overrides(self) -> list[Override]:
        if self.current_identifier is None:
            return []
        return [
            o for o in self.overrides if selector_matches(o.select_pattern, self.current_identifier)
        ]

    @contextlib.contextmanager
    def identifier(self, identifier: str | None) -> Generator[None, None, None]:
        self.current_identifier = identifier
        try:
            yield
        finally:
            self.current_identifier = None

    def get(
        self,
        name: str,
        *,
        env_plat: bool = True,
        option_format: OptionFormat | None = None,
        ignore_empty: bool = False,
        env_rule: InheritRule = InheritRule.NONE,
    ) -> str:
        """
        Get and return the value for the named option from environment,
        configuration file, or the default. If env_plat is False, then don't
        accept platform versions of the environment variable. If this is an
        array it will be merged with "sep" before returning. If it is a table,
        it will be formatted with "table['item']" using {k} and {v} and merged
        with "table['sep']". If sep is also given, it will be used for arrays
        inside the table (must match table['sep']). Empty variables will not
        override if ignore_empty is True.
        """

        if name not in self.default_options and name not in self.default_platform_options:
            msg = f"{name!r} must be in cibuildwheel/resources/defaults.toml file to be accessed."
            raise OptionsReaderError(msg)

        # Environment variable form
        envvar = f"CIBW_{name.upper().replace('-', '_')}"
        plat_envvar = f"{envvar}_{self.platform.upper()}"

        # get the option from the default, then the config file, then finally the environment.
        # platform-specific options are preferred, if they're allowed.
        return _resolve_cascade(
            (self.default_options.get(name), InheritRule.NONE),
            (self.default_platform_options.get(name), InheritRule.NONE),
            (self.config_options.get(name), InheritRule.NONE),
            (self.config_platform_options.get(name), InheritRule.NONE),
            *[
                (o.options.get(name), o.inherit.get(name, InheritRule.NONE))
                for o in self.active_config_overrides
            ],
            (self.env.get(envvar), env_rule),
            (self.env.get(plat_envvar) if env_plat else None, env_rule),
            ignore_empty=ignore_empty,
            option_format=option_format,
        )


class Options:
    pyproject_toml: dict[str, Any] | None

    def __init__(
        self,
        platform: PlatformName,
        command_line_arguments: CommandLineArguments,
        env: Mapping[str, str],
        defaults: bool = False,
    ):
        self.platform = platform
        self.command_line_arguments = command_line_arguments
        self.env = env
        self._defaults = defaults
        self._image_warnings = set[str]()

        self.reader = OptionsReader(
            None if defaults else self.config_file_path,
            platform=platform,
            env=env,
            disallow=DISALLOWED_OPTIONS,
        )

        self.package_dir = Path(command_line_arguments.package_dir)
        try:
            with self.package_dir.joinpath("pyproject.toml").open("rb") as f:
                self.pyproject_toml = tomllib.load(f)
        except FileNotFoundError:
            self.pyproject_toml = None

        # cache the build options method so repeated calls don't need to
        # resolve the options again
        self.build_options = functools.cache(self._compute_build_options)

    @functools.cached_property
    def config_file_path(self) -> Path | None:
        args = self.command_line_arguments

        if args.config_file:
            return Path(format_safe(args.config_file, package=args.package_dir))

        # return pyproject.toml, if it's available
        pyproject_toml_path = Path(args.package_dir) / "pyproject.toml"
        if pyproject_toml_path.exists():
            return pyproject_toml_path

        return None

    @functools.cached_property
    def package_requires_python_str(self) -> str | None:
        return get_requires_python_str(self.package_dir, self.pyproject_toml)

    @functools.cached_property
    def globals(self) -> GlobalOptions:
        args = self.command_line_arguments
        package_dir = args.package_dir
        output_dir = args.output_dir

        build_config = (
            self.reader.get("build", env_plat=False, option_format=ListFormat(sep=" ")) or "*"
        )
        skip_config = self.reader.get("skip", env_plat=False, option_format=ListFormat(sep=" "))
        test_skip = self.reader.get("test-skip", env_plat=False, option_format=ListFormat(sep=" "))

        allow_empty = args.allow_empty or strtobool(self.env.get("CIBW_ALLOW_EMPTY", "0"))

        enable_groups = self.reader.get(
            "enable", env_plat=False, option_format=ListFormat(sep=" "), env_rule=InheritRule.APPEND
        )
        try:
            enable = {
                *EnableGroup.parse_option_value(enable_groups),
                *EnableGroup.parse_option_value(" ".join(args.enable)),
            }
        except ValueError as e:
            msg = f"Failed to parse enable group. {e}. Valid group names are: {', '.join(g.value for g in EnableGroup)}"
            raise errors.ConfigurationError(msg) from e

        # This is not supported in tool.cibuildwheel, as it comes from a standard location.
        # Passing this in as an environment variable will override pyproject.toml, setup.cfg, or setup.py
        requires_python_str: str | None = (
            self.env.get("CIBW_PROJECT_REQUIRES_PYTHON") or self.package_requires_python_str
        )
        requires_python = None if requires_python_str is None else SpecifierSet(requires_python_str)

        archs_config_str = args.archs or self.reader.get(
            "archs", option_format=ListFormat(sep=" "), ignore_empty=True
        )
        architectures = Architecture.parse_config(archs_config_str, platform=self.platform)

        # Process `--only`
        if args.only:
            build_config = args.only
            skip_config = ""
            architectures = Architecture.all_archs(self.platform)
            enable |= EnableGroup.all_groups()
            if args.only.startswith("cp313t-"):
                enable.add(EnableGroup.CPythonFreeThreading)

        build_selector = BuildSelector(
            build_config=build_config,
            skip_config=skip_config,
            requires_python=requires_python,
            enable=frozenset(enable),
        )
        test_selector = TestSelector(skip_config=test_skip)

        return GlobalOptions(
            package_dir=package_dir,
            output_dir=output_dir,
            build_selector=build_selector,
            test_selector=test_selector,
            architectures=architectures,
            allow_empty=allow_empty,
        )

    def _check_pinned_image(self, value: str, pinned_images: Mapping[str, str]) -> None:
        error_set = {"manylinux1", "manylinux2010", "manylinux_2_24", "musllinux_1_1"}
        warning_set: set[str] = set()

        if value in error_set:
            msg = (
                f"cibuildwheel 3.x does not support the image {value!r}. Either upgrade to a "
                "supported image or continue using the image by pinning it directly with"
                " its full OCI registry '<name>{:<tag>|@<digest>}'."
            )
            raise errors.DeprecationError(msg)

        if value in warning_set and value not in self._image_warnings:
            self._image_warnings.add(value)
            msg = (
                f"Deprecated image {value!r}. This value will not work"
                " in a future version of cibuildwheel. Either upgrade to a supported"
                " image or continue using the deprecated image by pinning directly"
                f" to {pinned_images[value]!r}."
            )
            log.warning(msg)

    def _compute_build_options(self, identifier: str | None) -> BuildOptions:
        """
        Compute BuildOptions for a single run configuration. Normally accessed
        through the `build_options` method, which is the same but the result
        is cached.
        """

        with self.reader.identifier(identifier):
            before_all = self.reader.get("before-all", option_format=ListFormat(sep=" && "))

            environment_config = self.reader.get("environment", option_format=EnvironmentFormat())
            environment_pass = self.reader.get(
                "environment-pass", option_format=ListFormat(sep=" ")
            ).split()
            before_build = self.reader.get("before-build", option_format=ListFormat(sep=" && "))
            repair_command = self.reader.get(
                "repair-wheel-command", option_format=ListFormat(sep=" && ")
            )
            config_settings = self.reader.get(
                "config-settings", option_format=ShlexTableFormat(sep=" ", pair_sep="=")
            )

            test_command = self.reader.get("test-command", option_format=ListFormat(sep=" && "))
            before_test = self.reader.get("before-test", option_format=ListFormat(sep=" && "))
            xbuild_tools: list[str] | None = shlex.split(
                self.reader.get(
                    "xbuild-tools", option_format=ListFormat(sep=" ", quote=shlex.quote)
                )
            )
            # ["\u0000"] is a sentinel value used as a default, because TOML
            # doesn't have an explicit NULL value. If xbuild-tools is set to the
            # sentinel, it indicates that the user hasn't defined xbuild-tools
            # *at all* (not even an `xbuild-tools = []` definition).
            if xbuild_tools == ["\u0000"]:
                xbuild_tools = None

            test_sources = shlex.split(
                self.reader.get(
                    "test-sources", option_format=ListFormat(sep=" ", quote=shlex.quote)
                )
            )
            test_environment_config = self.reader.get(
                "test-environment", option_format=EnvironmentFormat()
            )
            try:
                test_environment = parse_environment(test_environment_config)
            except (EnvironmentParseError, ValueError) as e:
                msg = f"Malformed environment option {test_environment_config!r}"
                raise errors.ConfigurationError(msg) from e

            test_runtime_str = self.reader.get(
                "test-runtime",
                env_plat=False,
                option_format=ShlexTableFormat(sep="; ", pair_sep=":", allow_merge=False),
            )
            if not test_runtime_str:
                test_runtime = TestRuntimeConfig()
            else:
                try:
                    test_runtime = TestRuntimeConfig.from_config_string(test_runtime_str)
                except ValueError as e:
                    msg = f"Failed to parse test runtime config. {e}"
                    raise errors.ConfigurationError(msg) from e

            test_requires = self.reader.get(
                "test-requires", option_format=ListFormat(sep=" ")
            ).split()
            test_extras = self.reader.get("test-extras", option_format=ListFormat(sep=","))
            test_groups_str = self.reader.get("test-groups", option_format=ListFormat(sep=" "))
            test_groups = [x for x in test_groups_str.split() if x]
            test_requirements_from_groups = resolve_dependency_groups(
                self.pyproject_toml, *test_groups
            )
            build_verbosity_str = self.reader.get("build-verbosity")

            build_frontend_str = self.reader.get(
                "build-frontend",
                env_plat=False,
                option_format=ShlexTableFormat(sep="; ", pair_sep=":", allow_merge=False),
            )
            if not build_frontend_str or build_frontend_str == "default":
                build_frontend = BuildFrontendConfig("build")
            else:
                try:
                    build_frontend = BuildFrontendConfig.from_config_string(build_frontend_str)
                except ValueError as e:
                    msg = f"Failed to parse build frontend. {e}"
                    raise errors.ConfigurationError(msg) from e

            try:
                environment = parse_environment(environment_config)
            except (EnvironmentParseError, ValueError) as e:
                msg = f"Malformed environment option {environment_config!r}"
                raise errors.ConfigurationError(msg) from e

            # Pass through environment variables
            if self.platform == "linux":
                for env_var_name in reversed(environment_pass):
                    with contextlib.suppress(KeyError):
                        environment.add(env_var_name, self.env[env_var_name], prepend=True)

            dependency_versions_str = self.reader.get(
                "dependency-versions",
                env_plat=True,
                option_format=ShlexTableFormat(sep="; ", pair_sep=":", allow_merge=False),
            )
            try:
                dependency_constraints = DependencyConstraints.from_config_string(
                    dependency_versions_str
                )
            except (ValueError, OSError) as e:
                msg = f"Failed to parse dependency versions. {e}"
                raise errors.ConfigurationError(msg) from e

            if test_extras:
                test_extras = f"[{test_extras}]"

            try:
                build_verbosity = min(3, max(-3, int(build_verbosity_str)))
            except ValueError:
                build_verbosity = 0

            manylinux_images: dict[str, str] = {}
            musllinux_images: dict[str, str] = {}
            container_engine: OCIContainerEngineConfig | None = None

            if self.platform == "linux":
                all_pinned_container_images = _get_pinned_container_images()

                for build_platform in MANYLINUX_ARCHS:
                    pinned_images = all_pinned_container_images[build_platform]
                    config_value = self.reader.get(
                        f"manylinux-{build_platform}-image", ignore_empty=True
                    )
                    self._check_pinned_image(config_value, pinned_images)
                    if config_value in pinned_images:
                        image = pinned_images[config_value]
                    else:
                        image = config_value
                    manylinux_images[build_platform] = image

                for build_platform in MUSLLINUX_ARCHS:
                    pinned_images = all_pinned_container_images[build_platform]
                    config_value = self.reader.get(
                        f"musllinux-{build_platform}-image", ignore_empty=True
                    )
                    self._check_pinned_image(config_value, pinned_images)
                    if config_value in pinned_images:
                        image = pinned_images[config_value]
                    else:
                        image = config_value
                    musllinux_images[build_platform] = image

            container_engine_str = self.reader.get(
                "container-engine",
                option_format=ShlexTableFormat(sep="; ", pair_sep=":", allow_merge=False),
            )

            try:
                container_engine = OCIContainerEngineConfig.from_config_string(container_engine_str)
            except ValueError as e:
                msg = f"Failed to parse container config. {e}"
                raise errors.ConfigurationError(msg) from e

            pyodide_version = self.reader.get("pyodide-version", env_plat=False)

            return BuildOptions(
                globals=self.globals,
                test_command=test_command,
                test_sources=test_sources,
                test_environment=test_environment,
                test_runtime=test_runtime,
                test_requires=[*test_requires, *test_requirements_from_groups],
                test_extras=test_extras,
                test_groups=test_groups,
                before_test=before_test,
                before_build=before_build,
                before_all=before_all,
                build_verbosity=build_verbosity,
                xbuild_tools=xbuild_tools,
                repair_command=repair_command,
                environment=environment,
                dependency_constraints=dependency_constraints,
                manylinux_images=manylinux_images or None,
                musllinux_images=musllinux_images or None,
                build_frontend=build_frontend,
                config_settings=config_settings,
                container_engine=container_engine,
                pyodide_version=pyodide_version or None,
            )

    def check_for_invalid_configuration(self, identifiers: Iterable[str]) -> None:
        if self.platform in {"macos", "windows"}:
            before_all_values = {self.build_options(i).before_all for i in identifiers}

            if len(before_all_values) > 1:
                raise ValueError(
                    unwrap(
                        f"""
                        before_all cannot be set to multiple values. On macOS and Windows,
                        before_all is only run once, at the start of the build. before_all values
                        are: {before_all_values!r}
                        """
                    )
                )

    @functools.cached_property
    def defaults(self) -> Self:
        return self.__class__(
            platform=self.platform,
            command_line_arguments=CommandLineArguments.defaults(),
            env={},
            defaults=True,
        )

    def summary(self, identifiers: Iterable[str], skip_unset: bool = False) -> str:
        lines = []
        global_option_names = sorted(f.name for f in dataclasses.fields(self.globals))

        for option_name in global_option_names:
            option_value = getattr(self.globals, option_name)
            default_value = getattr(self.defaults.globals, option_name)
            line = self.option_summary(
                option_name, option_value, default_value, skip_unset=skip_unset
            )
            if line is not None:
                lines.append(line)

        build_options = self.build_options(identifier=None)
        build_options_defaults = self.defaults.build_options(identifier=None)
        build_options_for_identifier = {
            identifier: self.build_options(identifier) for identifier in identifiers
        }

        build_option_names = sorted(f.name for f in dataclasses.fields(build_options))

        for option_name in build_option_names:
            if option_name == "globals":
                continue

            option_value = getattr(build_options, option_name)
            default_value = getattr(build_options_defaults, option_name)
            overrides = {
                i: getattr(build_options_for_identifier[i], option_name) for i in identifiers
            }

            line = self.option_summary(
                option_name,
                option_value,
                default_value,
                overrides=overrides,
                skip_unset=skip_unset,
            )
            if line is not None:
                lines.append(line)

        return "\n".join(lines)

    def option_summary(
        self,
        option_name: str,
        option_value: Any,  # noqa: ANN401
        default_value: Any,  # noqa: ANN401
        overrides: Mapping[str, Any] | None = None,
        skip_unset: bool = False,
    ) -> str | None:
        """
        Return a summary of the option value, including any overrides, with
        ANSI 'dim' color if it's the default.
        """
        value_str = self.option_summary_value(option_value)
        default_value_str = self.option_summary_value(default_value)
        overrides_value_strs = {
            k: self.option_summary_value(v) for k, v in (overrides or {}).items()
        }
        # if the override value is the same as the non-overridden value, don't print it
        overrides_value_strs = {k: v for k, v in overrides_value_strs.items() if v != value_str}

        has_been_set = (value_str != default_value_str) or overrides_value_strs

        if skip_unset and not has_been_set:
            return None

        c = log.colors

        result = c.gray if not has_been_set else ""
        result += f"{option_name}: "

        if overrides_value_strs:
            overrides_groups = collections.defaultdict(list)
            for k, v in overrides_value_strs.items():
                overrides_groups[v].append(k)

            result += "\n  *: "
            result += self.indent_if_multiline(value_str, "    ")

            for override_value_str, identifiers in overrides_groups.items():
                result += f"\n  {', '.join(identifiers)}: "
                result += self.indent_if_multiline(override_value_str, "    ")
        else:
            result += self.indent_if_multiline(value_str, "  ")

        result += c.end

        return result

    @staticmethod
    def indent_if_multiline(value: str, indent: str) -> str:
        if "\n" in value:
            return "\n" + textwrap.indent(value.strip(), indent)
        else:
            return value

    @staticmethod
    def option_summary_value(option_value: Any) -> str:  # noqa: ANN401
        if hasattr(option_value, "options_summary"):
            option_value = option_value.options_summary()

        if isinstance(option_value, list):
            return "".join(f"{el}\n" for el in option_value)

        if isinstance(option_value, set):
            return ", ".join(str(el) for el in sorted(option_value))

        if isinstance(option_value, dict):
            return "".join(f"{k}: {v}\n" for k, v in option_value.items())

        return str(option_value)


def compute_options(
    platform: PlatformName,
    command_line_arguments: CommandLineArguments,
    env: Mapping[str, str],
) -> Options:
    return Options(platform=platform, command_line_arguments=command_line_arguments, env=env)


@functools.cache
def _get_pinned_container_images() -> Mapping[str, Mapping[str, str]]:
    """
    This looks like a dict of dicts, e.g.
    { 'x86_64': {'manylinux1': '...', 'manylinux2010': '...', 'manylinux2014': '...'},
      'i686': {'manylinux1': '...', 'manylinux2010': '...', 'manylinux2014': '...'},
      'pypy_x86_64': {'manylinux2010': '...' }
      ... }
    """
    all_pinned_images = configparser.ConfigParser()
    all_pinned_images.read(resources.PINNED_DOCKER_IMAGES)
    return all_pinned_images



import textwrap
from pathlib import Path

import pytest

from . import test_projects, utils

project_with_manylinux_symbols = test_projects.new_c_project(
    spam_c_top_level_add=textwrap.dedent(
        r"""
        #include <stdlib.h>

        #if defined(__GLIBC_PREREQ)
        #error "Must not run on a glibc linux environment"
        #endif
        """
    ),
    spam_c_function_add=textwrap.dedent(
        r"""
        sts = 0;
        """
    ),
)


@pytest.mark.parametrize(
    "musllinux_image",
    ["musllinux_1_2"],
)
@pytest.mark.usefixtures("docker_cleanup")
def test(musllinux_image: str, tmp_path: Path) -> None:
    if utils.get_platform() != "linux":
        pytest.skip("the container image test is only relevant to the linux build")

    project_dir = tmp_path / "project"
    project_with_manylinux_symbols.generate(project_dir)

    # build the wheels
    add_env = {
        "CIBW_SKIP": "*-manylinux*",
        "CIBW_MUSLLINUX_X86_64_IMAGE": musllinux_image,
        "CIBW_MUSLLINUX_I686_IMAGE": musllinux_image,
        "CIBW_MUSLLINUX_AARCH64_IMAGE": musllinux_image,
        "CIBW_MUSLLINUX_PPC64LE_IMAGE": musllinux_image,
        "CIBW_MUSLLINUX_S390X_IMAGE": musllinux_image,
        "CIBW_MUSLLINUX_ARMV7L_IMAGE": musllinux_image,
    }

    actual_wheels = utils.cibuildwheel_run(project_dir, add_env=add_env, single_python=True)
    expected_wheels = utils.expected_wheels(
        "spam",
        "0.1.0",
        manylinux_versions=[],
        musllinux_versions=[musllinux_image],
        single_python=True,
    )
    assert set(actual_wheels) == set(expected_wheels)



import os
import platform
import re
import sys
import typing
from collections.abc import Callable
from dataclasses import dataclass
from pathlib import Path
from shutil import rmtree, which
from subprocess import CalledProcessError
from textwrap import dedent
from zipfile import ZipFile

import pytest

from .test_projects import new_c_project
from .utils import cibuildwheel_run, expected_wheels

pytestmark = pytest.mark.android


CIBW_PLATFORM = os.environ.get("CIBW_PLATFORM", "android")
if CIBW_PLATFORM != "android":
    pytest.skip(f"{CIBW_PLATFORM=}", allow_module_level=True)

if (platform.system(), platform.machine()) not in [
    ("Linux", "x86_64"),
    ("Darwin", "arm64"),
    ("Darwin", "x86_64"),
]:
    pytest.skip(
        f"cibuildwheel does not support building Android wheels on "
        f"{platform.system()} {platform.machine()}",
        allow_module_level=True,
    )

# Azure Pipelines does not set the CI variable.
ci = any(key in os.environ for key in ["CI", "TF_BUILD"])

if "ANDROID_HOME" not in os.environ:
    msg = "ANDROID_HOME environment variable is not set"

    # Fail if we're on a CI service which is supposed to have the Android SDK
    # pre-installed; otherwise skip the module.
    if (
        ("CIRRUS_CI" in os.environ and platform.system() == "Darwin")
        or "GITHUB_ACTIONS" in os.environ
        or "TF_BUILD" in os.environ
    ):
        pytest.fail(msg)
    else:
        pytest.skip(msg, allow_module_level=True)

# Many CI services don't support running the Android emulator: see platforms.md.
supports_emulator = (not ci) or ("GITHUB_ACTIONS" in os.environ and platform.system() == "Linux")

T = typing.TypeVar("T", bound=Callable[..., typing.Any])


def needs_emulator(test: T) -> T:
    # All copies of the testbed app run on the same emulator with the same
    # application ID, so these tests must be run serially.
    test = pytest.mark.serial(test)

    if not supports_emulator:
        test = pytest.mark.skip("This CI platform doesn't support the emulator")(test)
    return test


@dataclass
class Architecture:
    linux_machine: str
    macos_machine: str
    android_abi: str


archs = [
    Architecture("aarch64", "arm64", "arm64_v8a"),
    Architecture("x86_64", "x86_64", "x86_64"),
]
native_arch = next(
    arch for arch in archs if platform.machine() in [arch.linux_machine, arch.macos_machine]
)


cp313_env = {
    "CIBW_PLATFORM": "android",
    "CIBW_BUILD": "cp313-*",
}


def test_android_home(tmp_path: Path, capfd: pytest.CaptureFixture[str]) -> None:
    new_c_project().generate(tmp_path)
    env = os.environ.copy()
    del env["ANDROID_HOME"]

    with pytest.raises(CalledProcessError):
        cibuildwheel_run(tmp_path, env={**env, **cp313_env})
    assert "ANDROID_HOME environment variable is not set" in capfd.readouterr().err


# android-env.sh may need to install the NDK, and it isn't safe to do that multiple
# times in parallel. So make sure there's at least one test which gets as far as doing
# a build, which is marked as serial so it will run before the parallel tests, but isn't
# marked as needs_emulator so it will run on all CI platforms.
@pytest.mark.serial
def test_expected_wheels(tmp_path: Path, spam_env: dict[str, str]) -> None:
    # Since this test covers all Python versions, check the cross venv.
    test_module = "_cross_venv_test_android"
    project = new_c_project(setup_py_add=f"import {test_module}")
    project.files[f"{test_module}.py"] = (Path(__file__).parent / f"{test_module}.py").read_text()
    project.generate(tmp_path)

    # Build wheels for all Python versions on the current architecture.
    del spam_env["CIBW_BUILD"]
    if not supports_emulator:
        del spam_env["CIBW_TEST_COMMAND"]

    wheels = cibuildwheel_run(tmp_path, add_env=spam_env)
    assert wheels == expected_wheels(
        "spam", "0.1.0", platform="android", machine_arch=native_arch.android_abi
    )


@needs_emulator
def test_frontend_good(tmp_path: Path, build_frontend_env: dict[str, str]) -> None:
    new_c_project().generate(tmp_path)
    wheels = cibuildwheel_run(
        tmp_path,
        add_env={**cp313_env, **build_frontend_env, "CIBW_TEST_COMMAND": "python -m site"},
    )
    assert wheels == [f"spam-0.1.0-cp313-cp313-android_21_{native_arch.android_abi}.whl"]


@pytest.mark.parametrize("frontend", ["pip"])
def test_frontend_bad(frontend: str, tmp_path: Path, capfd: pytest.CaptureFixture[str]) -> None:
    new_c_project().generate(tmp_path)
    with pytest.raises(CalledProcessError):
        cibuildwheel_run(
            tmp_path,
            add_env={**cp313_env, "CIBW_BUILD_FRONTEND": frontend},
        )
    assert "Android requires the build frontend to be 'build'" in capfd.readouterr().err


@needs_emulator
def test_archs(tmp_path: Path, capfd: pytest.CaptureFixture[str]) -> None:
    new_c_project().generate(tmp_path)

    # Build all architectures while checking the handling of the `before` commands.
    command_pattern = 'echo "Hello from {0}, package={{package}}, python=$(which python)"'
    output_pattern = (
        f"Hello from {{0}}, package={tmp_path}, python=/.+/cp313-android_{{1}}/venv/bin/python"
    )

    wheels = cibuildwheel_run(
        tmp_path,
        add_env={
            **cp313_env,
            "CIBW_ARCHS": "all",
            "CIBW_BEFORE_ALL": "echo 'Hello from before_all'",
            "CIBW_BEFORE_BUILD": command_pattern.format("before_build"),
            "CIBW_BEFORE_TEST": command_pattern.format("before_test"),
            "CIBW_TEST_COMMAND": (
                "python -c 'import platform; print(f\"Hello from {platform.machine()}\")'"
            ),
        },
    )
    assert wheels == [f"spam-0.1.0-cp313-cp313-android_21_{arch.android_abi}.whl" for arch in archs]

    stdout, stderr = capfd.readouterr()
    lines = (line for line in stdout.splitlines() if line.startswith("Hello from"))
    assert next(lines) == "Hello from before_all"

    # All architectures should be built, but only the native architecture should run tests.
    for arch in archs:
        abi = arch.android_abi
        assert re.fullmatch(output_pattern.format("before_build", abi), next(lines))
        if arch == native_arch:
            assert re.fullmatch(output_pattern.format("before_test", abi), next(lines))
            assert next(lines) == f"Hello from {arch.linux_machine}"
        else:
            assert (
                f"Skipping tests for {arch.android_abi}, as the build machine "
                f"only supports {native_arch.android_abi}"
            ) in stderr

    try:
        line = next(lines)
    except StopIteration:
        pass
    else:
        pytest.fail(f"Unexpected line: {line!r}")


def test_build_requires(tmp_path: Path, capfd: pytest.CaptureFixture[str]) -> None:
    # Build-time requirements should be installed for the build platform, not for Android. Prove
    # this by installing some non-pure-Python requirements and using them in setup.py.
    #
    # setup_requires is installed via ProjectBuilder.get_requires_for_build.
    project = new_c_project(
        setup_py_setup_args_add="setup_requires=['cmake==3.31.4']",
        setup_py_add=dedent(
            """\
            if "egg_info" not in sys.argv:
                import subprocess
                subprocess.run(["cmake", "--version"], check=True)

                from bitarray import bitarray
                print(f"{bitarray('10110').count()=}")
            """
        ),
    )

    # [build_system] requires is installed via ProjectBuilder.build_system_requires.
    project.files["pyproject.toml"] = dedent(
        """\
        [build-system]
        requires = ["setuptools", "wheel", "bitarray==3.3.2"]
        """
    )

    project.generate(tmp_path)
    cibuildwheel_run(tmp_path, add_env={**cp313_env})

    # Test for a specific version to minimize the chance that we ran a system cmake.
    stdout = capfd.readouterr().out
    assert "cmake version 3.31.4" in stdout
    assert "bitarray('10110').count()=3" in stdout


@pytest.fixture
def spam_env(tmp_path: Path) -> dict[str, str]:
    project = new_c_project()
    project.files["test_spam.py"] = dedent(
        """\
        import spam

        def test_spam() -> None:
            assert spam.filter("ham")
            assert not spam.filter("spam")
            print("Spam test passed")
        """
    )
    project.files["test_empty.py"] = dedent(
        """\
        def test_empty() -> None:
            pass
        """
    )

    project.generate(tmp_path)

    return {
        **cp313_env,
        "CIBW_TEST_SOURCES": "test_spam.py test_empty.py",
        "CIBW_TEST_REQUIRES": "pytest==8.3.5",
        "CIBW_TEST_COMMAND": "python -m pytest",
    }


@needs_emulator
@pytest.mark.parametrize(
    ("command", "expected_output"),
    [
        ("python3 -c 'import test_spam; test_spam.test_spam()'", "Spam test passed"),
        ("python -m pytest", "=== 2 passed in "),
        ("python -m pytest test_spam.py", "=== 1 passed in "),
        ("pytest test_spam.py", "=== 1 passed in "),
    ],
)
def test_test_command_good(
    command: str,
    expected_output: str,
    tmp_path: Path,
    spam_env: dict[str, str],
    capfd: pytest.CaptureFixture[str],
) -> None:
    cibuildwheel_run(tmp_path, add_env={**spam_env, "CIBW_TEST_COMMAND": command})
    stdout, stderr = capfd.readouterr()
    assert expected_output in stdout

    if not command.startswith("python"):
        assert (
            f"Test command {command!r} is not supported on Android. cibuildwheel "
            "will try to execute it as if it started with 'python -m'."
        ) in stderr


BAD_FORMAT_ERROR = (
    "Test command '{}' is not supported on Android. "
    "Command must begin with 'python' or 'python3', and contain '-m' or '-c'."
)
BAD_PLACEHOLDER_ERROR = (
    "Test command '{}' with a '{{project}}' or '{{package}}' placeholder "
    "is not supported on Android"
)


@needs_emulator
@pytest.mark.parametrize(
    ("command", "expected_output"),
    [
        # Build-time failure
        ("./test_spam.py", BAD_FORMAT_ERROR.format("./test_spam.py")),
        ("python test_spam.py", BAD_FORMAT_ERROR.format("python test_spam.py")),
        ("pytest {project}", BAD_PLACEHOLDER_ERROR.format("pytest {project}")),
        ("pytest {package}", BAD_PLACEHOLDER_ERROR.format("pytest {package}")),
        # Runtime failure
        ("pytest test_ham.py", "not found: test_ham.py"),
    ],
)
def test_test_command_bad(
    command: str,
    expected_output: str,
    tmp_path: Path,
    spam_env: dict[str, str],
    capfd: pytest.CaptureFixture[str],
) -> None:
    with pytest.raises(CalledProcessError):
        cibuildwheel_run(tmp_path, add_env={**spam_env, "CIBW_TEST_COMMAND": command})
    assert expected_output in capfd.readouterr().err


@needs_emulator
@pytest.mark.parametrize(
    ("options", "expected"),
    [
        ("", 0),
        ("-E", 1),
    ],
)
def test_test_command_python_options(
    options: str, expected: int, tmp_path: Path, capfd: pytest.CaptureFixture[str]
) -> None:
    project = new_c_project()
    project.generate(tmp_path)

    command = 'import sys; print(f"{sys.flags.ignore_environment=}")'
    cibuildwheel_run(
        tmp_path,
        add_env={
            **cp313_env,
            "CIBW_TEST_COMMAND": f"python {options} -c '{command}'",
        },
    )
    assert f"sys.flags.ignore_environment={expected}" in capfd.readouterr().out


@needs_emulator
def test_package_subdir(
    tmp_path: Path, spam_env: dict[str, str], capfd: pytest.CaptureFixture[str]
) -> None:
    spam_paths = list(tmp_path.iterdir())
    package_dir = tmp_path / "package"
    package_dir.mkdir()
    for path in spam_paths:
        path.rename(package_dir / path.name)

    spam_env["CIBW_TEST_SOURCES"] = " ".join(
        f"package/{path}" for path in spam_env["CIBW_TEST_SOURCES"].split()
    )
    cibuildwheel_run(tmp_path, package_dir, add_env=spam_env)
    assert "=== 2 passed in " in capfd.readouterr().out


@needs_emulator
def test_no_test_sources(tmp_path: Path, capfd: pytest.CaptureFixture[str]) -> None:
    new_c_project().generate(tmp_path)
    with pytest.raises(CalledProcessError):
        cibuildwheel_run(
            tmp_path,
            add_env={**cp313_env, "CIBW_TEST_COMMAND": "python -m unittest discover"},
        )
    assert (
        "On this platform, you must copy your test files to the testbed app by "
        "setting the `test-sources` option"
    ) in capfd.readouterr().err


@needs_emulator
def test_environment_markers(tmp_path: Path) -> None:
    project = new_c_project()
    test_filename = "test_environment_markers.py"
    project.files[test_filename] = dedent(
        """\
        import pytest

        def test_android() -> None:
            import certifi

        def test_not_android() -> None:
            try:
                import platformdirs
            except ImportError:
                pass
            else:
                pytest.fail("`platformdirs` should not have been installed")
        """
    )
    project.generate(tmp_path)

    cibuildwheel_run(
        tmp_path,
        add_env={
            **cp313_env,
            "CIBW_TEST_COMMAND": f"python -m pytest {test_filename}",
            "CIBW_TEST_SOURCES": test_filename,
            "CIBW_TEST_REQUIRES": "pytest certifi;sys_platform=='android' platformdirs;sys_platform!='android'",
        },
    )


@needs_emulator
def test_verbosity(tmp_path: Path, capfd: pytest.CaptureFixture[str]) -> None:
    new_c_project().generate(tmp_path)
    test_env = {
        **cp313_env,
        "CIBW_TEST_COMMAND": """python -c 'print("Hello world")'""",
    }
    verbose_lines = [
        "> Task :app:packageDebug",  # Gradle
        "I/TestRunner: run started: 1 tests",  # Logcat
    ]

    cibuildwheel_run(tmp_path, add_env=test_env)
    stdout = capfd.readouterr().out
    for line in verbose_lines:
        assert line not in stdout

    cibuildwheel_run(
        tmp_path,
        add_env={**test_env, "CIBW_BUILD_VERBOSITY": "1"},
    )
    stdout = capfd.readouterr().out
    for line in verbose_lines:
        assert line in stdout


@needs_emulator
def test_api_level(tmp_path: Path, capfd: pytest.CaptureFixture[str]) -> None:
    project = new_c_project()
    project.files["pyproject.toml"] = dedent(
        """\
        [build-system]
        requires = ["setuptools"]

        [tool.cibuildwheel]
        android.environment.ANDROID_API_LEVEL = "33"
        android.environment.PIP_EXTRA_INDEX_URL = "https://chaquo.com/pypi-13.1"
        """
    )
    project.generate(tmp_path)

    wheels = cibuildwheel_run(
        tmp_path,
        add_env={
            **cp313_env,
            # Verify that Android dependencies can be installed from the Chaquopy repository, and
            # that wheels tagged with an older version of Android (in this case 24) are still
            # accepted.
            "CIBW_TEST_REQUIRES": "bitarray==3.0.0",
            "CIBW_TEST_COMMAND": (
                "python -c 'from bitarray import bitarray; print(~bitarray(\"01100\"))'"
            ),
        },
    )
    assert wheels == [f"spam-0.1.0-cp313-cp313-android_33_{native_arch.android_abi}.whl"]
    assert "bitarray('10011')" in capfd.readouterr().out


@needs_emulator
def test_libcxx(tmp_path: Path, capfd: pytest.CaptureFixture[str]) -> None:
    project_dir = tmp_path / "project"
    output_dir = tmp_path / "output"

    # cibuildwheel should be able to run `patchelf` and `wheel` even when its
    # environment's `bin` directory is not on the PATH.
    non_venv_path = ":".join(
        item for item in os.environ["PATH"].split(":") if Path(item) != Path(sys.executable).parent
    )

    # A C++ package should include libc++, and the extension module should be able to
    # find it using DT_RUNPATH.
    new_c_project(setup_py_extension_args_add="language='c++'").generate(project_dir)
    script = 'import spam; print(", ".join(f"{s}: {spam.filter(s)}" for s in ["ham", "spam"]))'
    cp313_test_env = {
        **cp313_env,
        "CIBW_TEST_COMMAND": f"python -c '{script}'",
        "PATH": non_venv_path,
    }

    # Including external libraries requires API level 24.
    with pytest.raises(CalledProcessError):
        cibuildwheel_run(project_dir, add_env=cp313_test_env, output_dir=output_dir)
    assert "libc++_shared.so requires ANDROID_API_LEVEL to be at least 24" in capfd.readouterr().err

    wheels = cibuildwheel_run(
        project_dir,
        add_env={**cp313_test_env, "ANDROID_API_LEVEL": "24"},
        output_dir=output_dir,
    )
    assert len(wheels) == 1
    names = ZipFile(output_dir / wheels[0]).namelist()
    libcxx_names = [
        name for name in names if re.fullmatch(r"spam\.libs/libc\+\+_shared-[0-9a-f]{8}\.so", name)
    ]
    assert len(libcxx_names) == 1
    assert "ham: 1, spam: 0" in capfd.readouterr().out

    # A C package should not include libc++.
    rmtree(project_dir)
    rmtree(output_dir)
    new_c_project().generate(project_dir)
    wheels = cibuildwheel_run(project_dir, add_env=cp313_env, output_dir=output_dir)
    assert len(wheels) == 1
    for name in ZipFile(output_dir / wheels[0]).namelist():
        assert ".libs" not in name


@needs_emulator
def test_setuptools_rust(tmp_path: Path, capfd: pytest.CaptureFixture[str]) -> None:
    """
    Test Android cross-compilation using the setuptools-rust toolchain.
    """
    if not which("rustup") or not which("cargo"):
        pytest.skip("rustup and cargo are required for this test")

    # Initialize a basic project and override files for setuptools-rust
    project = new_c_project()
    project.files["pyproject.toml"] = dedent(
        """\
       [build-system]
       requires = ["setuptools", "wheel", "setuptools-rust"]
       """
    )
    project.files["setup.py"] = dedent(
        """\
       from setuptools import setup
       from setuptools_rust import Binding, RustExtension

       setup(
           name="spam",
           version="0.1.0",
           rust_extensions=[RustExtension("spam.rust_ext", binding=Binding.PyO3)],
           zip_safe=False,
       )
       """
    )
    project.files["Cargo.toml"] = dedent(
        """\
       [package]
       name = "rust_ext"
       version = "0.1.0"
       edition = "2021"

       [lib]
       name = "rust_ext"
       crate-type = ["cdylib"]

       [dependencies]
       pyo3 = { version = "0.23.3", features = ["extension-module"] }
       """
    )

    # Create the Rust source directory and file
    (tmp_path / "src").mkdir()
    project.files["src/lib.rs"] = dedent(
        """\
       use pyo3::prelude::*;

       #[pyfunction]
       fn hello() -> PyResult<String> {
           Ok("Hello from Rust via setuptools-rust!".to_string())
       }

       #[pymodule]
       fn rust_ext(_py: Python, m: &Bound<'_, PyModule>) -> PyResult<()> {
           m.add_function(wrap_pyfunction!(hello, m)?)?;
           Ok(())
       }
       """
    )

    project.generate(tmp_path)

    # Command to verify the built extension on the Android emulator
    test_command = "python -c 'import spam.rust_ext; print(spam.rust_ext.hello())'"

    # Run cibuildwheel to build and test the wheel
    wheels = cibuildwheel_run(
        tmp_path,
        add_env={
            **cp313_env,
            "CIBW_TEST_COMMAND": test_command,
            "CIBW_BUILD_VERBOSITY": "1",
        },
    )

    # Verification
    assert len(wheels) == 1
    stdout = capfd.readouterr().out
    assert "Hello from Rust via setuptools-rust!" in stdout


@needs_emulator
def test_maturin(tmp_path: Path, capfd: pytest.CaptureFixture[str]) -> None:
    """
    Test Android cross-compilation using the maturin backend.
    """
    if not which("rustup") or not which("cargo"):
        pytest.skip("rustup and cargo are required for this test")

    project = new_c_project()
    project.files["pyproject.toml"] = dedent(
        """\
        [build-system]
        # maturin >= 1.11.0 is required for reliable Android cross-compilation support.
        requires = ["maturin>=1.11.0,<2.0"]
        build-backend = "maturin"

        [project]
        name = "spam"
        version = "0.1.0"
        """
    )
    project.files["Cargo.toml"] = dedent(
        """\
        [package]
        name = "spam"
        version = "0.1.0"
        edition = "2021"

        [lib]
        name = "spam"
        crate-type = ["cdylib"]

        [dependencies]
        pyo3 = { version = "0.23.3", features = ["extension-module"] }
        """
    )

    (tmp_path / "src").mkdir()
    project.files["src/lib.rs"] = dedent(
        """\
        use pyo3::prelude::*;

        #[pyfunction]
        fn hello() -> PyResult<String> {
            Ok("Hello from Rust via maturin!".to_string())
        }

        #[pymodule]
        fn spam(_py: Python, m: &Bound<'_, PyModule>) -> PyResult<()> {
            m.add_function(wrap_pyfunction!(hello, m)?)?;
            Ok(())
        }
        """
    )

    project.generate(tmp_path)
    (tmp_path / "setup.py").unlink()

    test_command = "python -c 'import spam; print(spam.hello())'"

    wheels = cibuildwheel_run(
        tmp_path,
        add_env={
            **cp313_env,
            "CIBW_TEST_COMMAND": test_command,
            "CIBW_BUILD_VERBOSITY": "1",
        },
    )

    assert len(wheels) == 1
    stdout = capfd.readouterr().out
    assert "Hello from Rust via maturin!" in stdout



import dataclasses
import itertools
from enum import StrEnum
from fnmatch import fnmatch
from typing import Self

import bracex
from packaging.specifiers import SpecifierSet
from packaging.version import Version


def selector_matches(patterns: str, string: str) -> bool:
    """
    Returns True if `string` is matched by any of the wildcard patterns in
    `patterns`.

    Matching is according to fnmatch, but with shell-like curly brace
    expansion. For example, 'cp{36,37}-*' would match either of 'cp36-*' or
    'cp37-*'.
    """

    patterns_list = patterns.split()
    expanded_patterns = itertools.chain.from_iterable(bracex.expand(p) for p in patterns_list)
    return any(fnmatch(string, pat) for pat in expanded_patterns)


class EnableGroup(StrEnum):
    """
    Groups of build selectors that are not enabled by default.
    """

    CPythonFreeThreading = "cpython-freethreading"
    CPythonPrerelease = "cpython-prerelease"
    GraalPy = "graalpy"
    PyPy = "pypy"
    PyPyEoL = "pypy-eol"
    PyodidePrerelease = "pyodide-prerelease"

    @classmethod
    def all_groups(cls) -> frozenset[Self]:
        # see https://github.com/python/mypy/issues/20434 for the reason to type: ignore[arg-type]
        return frozenset(set(cls) - {cls.CPythonFreeThreading})  # type: ignore[arg-type]

    @classmethod
    def parse_option_value(cls, value: str) -> frozenset[Self]:
        """
        Parses a string of space-separated values into a set of EnableGroup
        members. The string may contain group names or "all".
        """
        result = set()
        for group in value.strip().split():
            if group == "all":
                return cls.all_groups()
            try:
                result.add(cls(group))
            except ValueError:
                msg = f"Unknown enable group: {group}"
                raise ValueError(msg) from None
        return frozenset(result)


@dataclasses.dataclass(frozen=True, kw_only=True)
class BuildSelector:
    """
    This class holds a set of build/skip patterns. You call an instance with a
    build identifier, and it returns True if that identifier should be
    included. Only call this on valid identifiers, ones that have at least 2
    numeric digits before the first dash.
    """

    build_config: str
    skip_config: str
    requires_python: SpecifierSet | None = None
    enable: frozenset[EnableGroup] = frozenset()

    def __call__(self, build_id: str) -> bool:
        # Filter build selectors by python_requires if set
        if self.requires_python is not None:
            py_ver_str = build_id.split("-", maxsplit=1)[0].split("_", maxsplit=1)[0]
            py_ver_str = py_ver_str.removesuffix("t")
            major = int(py_ver_str[2])
            minor = int(py_ver_str[3:])
            version = Version(f"{major}.{minor}.99")
            if not self.requires_python.contains(version):
                return False

        # filter out groups that are not enabled
        if EnableGroup.CPythonFreeThreading not in self.enable and fnmatch(build_id, "cp313t-*"):
            return False
        if EnableGroup.CPythonPrerelease not in self.enable and fnmatch(build_id, "cp315*"):
            return False
        is_pypy_eol = fnmatch(build_id, "pp3?-*") or fnmatch(build_id, "pp310-*")
        is_pypy = fnmatch(build_id, "pp*") and not is_pypy_eol
        if EnableGroup.PyPy not in self.enable and is_pypy:
            return False
        if EnableGroup.PyPyEoL not in self.enable and is_pypy_eol:
            return False
        if EnableGroup.GraalPy not in self.enable and fnmatch(build_id, "gp*"):
            return False
        # TODO: Re-enable this when we have Pyodide prereleases again (e.g., 0.29.0a1+)
        # Python 3.13 support became stable in Pyodide 0.28.0, so it no longer needs a prerelease
        # flag.
        # Also update Pyodide tests in unit_test/build_selector_test.py accordingly.
        # When re-enabling, update the pattern to match the experimental Python version in case
        # it is bumped to Python 3.14 (likely cp314-pyodide_* but could remain as 3.13 as well).
        # This depends on the CPython version being used in the Pyodide runtime at the time.
        # if EnableGroup.PyodidePrerelease not in self.enable and fnmatch(
        #     build_id, "cp313-pyodide_*"
        # ):
        #     return False

        should_build = selector_matches(self.build_config, build_id)
        should_skip = selector_matches(self.skip_config, build_id)

        return should_build and not should_skip

    def options_summary(self) -> dict[str, str | list[str]]:
        return {
            "build_config": self.build_config,
            "skip_config": self.skip_config,
            "requires_python": str(self.requires_python),
            "enable": sorted(group.value for group in self.enable),
        }


@dataclasses.dataclass(frozen=True, kw_only=True)
class TestSelector:
    """
    A build selector that can only skip tests according to a skip pattern.
    """

    skip_config: str

    def __call__(self, build_id: str) -> bool:
        should_skip = selector_matches(self.skip_config, build_id)
        return not should_skip

    def options_summary(self) -> dict[str, str]:
        return {"skip_config": self.skip_config}



import itertools
import subprocess
from pathlib import Path

import pytest

from . import test_projects, utils

project_with_a_test = test_projects.new_c_project()

project_with_a_test.files["test/spam_test.py"] = r"""
import spam

def test_spam() -> None:
    assert spam.filter("spam") == 0
    assert spam.filter("ham") != 0
"""


def test(tmp_path: Path, request: pytest.FixtureRequest) -> None:
    archs = request.config.getoption("--run-emulation")
    if archs is None:
        pytest.skip("needs --run-emulation option to run")

    if archs == "all":
        archs = " ".join(utils.EMULATED_ARCHS)

    project_dir = tmp_path / "project"
    project_with_a_test.generate(project_dir)

    # build and test the wheels
    actual_wheels = utils.cibuildwheel_run(
        project_dir,
        add_env={
            "CIBW_TEST_REQUIRES": "pytest",
            "CIBW_TEST_COMMAND": "pytest {project}/test",
            "CIBW_ARCHS": archs,
        },
    )

    # also check that we got the right wheels
    expected_wheels = itertools.chain.from_iterable(
        utils.expected_wheels("spam", "0.1.0", machine_arch=arch, single_arch=True)
        for arch in archs.split(" ")
    )
    assert set(actual_wheels) == set(expected_wheels)


def test_setting_arch_on_other_platforms(tmp_path: Path, capfd: pytest.CaptureFixture[str]) -> None:
    if utils.get_platform() == "linux":
        pytest.skip("this test checks the behaviour on platforms other than linux")

    project_dir = tmp_path / "project"
    project_with_a_test.generate(project_dir)

    # build and test the wheels
    with pytest.raises(subprocess.CalledProcessError):
        utils.cibuildwheel_run(
            project_dir,
            add_env={
                "CIBW_ARCHS": "aarch64",
            },
        )

    captured = capfd.readouterr()
    assert "Invalid archs option" in captured.err



from pathlib import Path

from . import test_projects, utils

before_test_project = test_projects.new_c_project()
before_test_project.files["test/spam_test.py"] = r"""
import sys
import os
from pathlib import Path
from unittest import TestCase

PROJECT_DIR = Path(__file__).parent.parent.resolve()


class TestBeforeTest(TestCase):
    def test_version(self) -> None:
        # assert that the Python version as written to pythonversion_bt.txt in the CIBW_BEFORE_TEST step
        # is the same one as is currently running.
        # because of use symlinks in MacOS run this test is also need
        stored_version = PROJECT_DIR.joinpath('pythonversion_bt.txt').read_text()
        print('stored_version', stored_version)
        print('sys.version', sys.version)
        assert stored_version == sys.version

    def test_prefix(self) -> None:
        # check that the prefix also was written
        stored_prefix = PROJECT_DIR.joinpath('pythonprefix_bt.txt').read_text()
        print('stored_prefix', stored_prefix)
        print('sys.prefix', sys.prefix)
        #  Works around path-comparison bugs caused by short-paths on Windows e.g.
        #  vssadm~1 instead of vssadministrator

        assert os.path.samefile(stored_prefix, sys.prefix)
"""


def test(tmp_path: Path, build_frontend_env: dict[str, str]) -> None:
    project_dir = tmp_path / "project"
    before_test_project.generate(project_dir)
    test_project_dir = project_dir / "dependency"
    test_projects.new_c_project().generate(test_project_dir)

    before_test_steps = [
        '''python -c "import pathlib, sys; pathlib.Path('{project}/pythonversion_bt.txt').write_text(sys.version)"''',
        '''python -c "import pathlib, sys; pathlib.Path('{project}/pythonprefix_bt.txt').write_text(sys.prefix)"''',
    ]

    if utils.get_platform() == "pyodide":
        before_test_steps.extend(
            ["pyodide build {project}/dependency", "pip install --find-links dist/ spam"]
        )
    elif build_frontend_env["CIBW_BUILD_FRONTEND"] in {"pip", "build"}:
        before_test_steps.append("python -m pip install {project}/dependency")

    before_test = " && ".join(before_test_steps)

    # build the wheels
    actual_wheels = utils.cibuildwheel_run(
        project_dir,
        add_env={
            # write python version information to a temporary file, this is
            # checked in setup.py
            "CIBW_BEFORE_TEST": before_test,
            "CIBW_TEST_REQUIRES": "pytest",
            # the 'false ||' bit is to ensure this command runs in a shell on
            # mac/linux.
            "CIBW_TEST_COMMAND": f"false || {utils.invoke_pytest()} {{project}}/test",
            "CIBW_TEST_COMMAND_WINDOWS": "pytest {project}/test",
            **build_frontend_env,
        },
    )

    # also check that we got the right wheels
    expected_wheels = utils.expected_wheels("spam", "0.1.0")
    assert set(actual_wheels) == set(expected_wheels)
